# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime

import mock
import pytest

from hikari import channels, urls
from hikari import files
from hikari import permissions
from hikari import snowflakes
from hikari import users
from hikari import webhooks
from hikari.internal import routes
from tests.hikari import hikari_test_helpers


@pytest.fixture
def mock_app():
    return mock.Mock()


class TestChannelFollow:
    @pytest.mark.asyncio
    async def test_fetch_channel(self, mock_app):
        mock_channel = mock.Mock(spec=channels.GuildNewsChannel)
        mock_app.rest.fetch_channel = mock.AsyncMock(return_value=mock_channel)
        follow = channels.ChannelFollow(
            channel_id=snowflakes.Snowflake(9459234123), app=mock_app, webhook_id=snowflakes.Snowflake(3123123)
        )

        result = await follow.fetch_channel()

        assert result is mock_channel
        mock_app.rest.fetch_channel.assert_awaited_once_with(9459234123)

    @pytest.mark.asyncio
    async def test_fetch_webhook(self, mock_app):
        mock_app.rest.fetch_webhook = mock.AsyncMock(return_value=mock.Mock(webhooks.ChannelFollowerWebhook))
        follow = channels.ChannelFollow(
            webhook_id=snowflakes.Snowflake(54123123), app=mock_app, channel_id=snowflakes.Snowflake(94949494)
        )

        result = await follow.fetch_webhook()

        assert result is mock_app.rest.fetch_webhook.return_value
        mock_app.rest.fetch_webhook.assert_awaited_once_with(54123123)

    def test_get_channel(self, mock_app):
        mock_channel = mock.Mock(spec=channels.GuildNewsChannel)
        mock_app.cache.get_guild_channel = mock.Mock(return_value=mock_channel)
        follow = channels.ChannelFollow(
            webhook_id=snowflakes.Snowflake(993883), app=mock_app, channel_id=snowflakes.Snowflake(696969)
        )

        result = follow.get_channel()

        assert result is mock_channel
        mock_app.cache.get_guild_channel.assert_called_once_with(696969)

    def test_get_channel_when_no_cache_trait(self):
        follow = channels.ChannelFollow(
            webhook_id=snowflakes.Snowflake(993883), app=object(), channel_id=snowflakes.Snowflake(696969)
        )

        assert follow.get_channel() is None


class TestPermissionOverwrite:
    def test_unset(self):
        overwrite = channels.PermissionOverwrite(
            type=channels.PermissionOverwriteType.MEMBER, id=snowflakes.Snowflake(1234321)
        )
        overwrite.allow = permissions.Permissions.CREATE_INSTANT_INVITE
        overwrite.deny = permissions.Permissions.CHANGE_NICKNAME
        assert overwrite.unset == permissions.Permissions(-67108866)


class TestPartialChannel:
    @pytest.fixture
    def model(self, mock_app):
        return hikari_test_helpers.mock_class_namespace(channels.PartialChannel, rename_impl_=False)(
            app=mock_app, id=snowflakes.Snowflake(1234567), name="foo", type=channels.ChannelType.GUILD_NEWS
        )

    def test_str_operator(self, model):
        assert str(model) == "foo"

    def test_str_operator_when_name_is_None(self, model):
        model.name = None
        assert str(model) == "Unnamed PartialChannel ID 1234567"

    def test_mention_property(self, model):
        assert model.mention == "<#1234567>"

    @pytest.mark.asyncio
    async def test_delete(self, model):
        model.app.rest.delete_channel = mock.AsyncMock()

        assert await model.delete() is model.app.rest.delete_channel.return_value

        model.app.rest.delete_channel.assert_called_once_with(1234567)


class TestDMChannel:
    @pytest.fixture
    def model(self, mock_app):
        return channels.DMChannel(
            id=snowflakes.Snowflake(12345),
            name="steve",
            type=channels.ChannelType.DM,
            last_message_id=snowflakes.Snowflake(12345),
            recipient=mock.Mock(spec_set=users.UserImpl, __str__=mock.Mock(return_value="snoop#0420")),
            app=mock_app,
        )

    def test_str_operator(self, model):
        assert str(model) == "DMChannel with: snoop#0420"

    def test_shard_id(self, model):
        assert model.shard_id == 0


class TestGroupDMChannel:
    @pytest.fixture
    def model(self, mock_app):
        return channels.GroupDMChannel(
            app=mock_app,
            id=snowflakes.Snowflake(136134),
            name="super cool group dm",
            type=channels.ChannelType.DM,
            last_message_id=snowflakes.Snowflake(3232),
            owner_id=snowflakes.Snowflake(1066),
            icon_hash="1a2b3c",
            nicknames={snowflakes.Snowflake(1): "person 1", snowflakes.Snowflake(2): "person 2"},
            recipients={
                snowflakes.Snowflake(1): mock.Mock(spec_set=users.User, __str__=mock.Mock(return_value="snoop#0420")),
                snowflakes.Snowflake(2): mock.Mock(spec_set=users.User, __str__=mock.Mock(return_value="yeet#1012")),
                snowflakes.Snowflake(3): mock.Mock(spec_set=users.User, __str__=mock.Mock(return_value="nice#6969")),
            },
            application_id=None,
        )

    def test_str_operator(self, model):
        assert str(model) == "super cool group dm"

    def test_str_operator_when_name_is_None(self, model):
        model.name = None
        assert str(model) == "GroupDMChannel with: snoop#0420, yeet#1012, nice#6969"

    def test_make_icon_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_CHANNEL_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, channel_id=136134, hash="1a2b3c", size=4096, file_format="JPEG", lossless=True
        )

    def test_icon_url(self):
        channel = hikari_test_helpers.mock_class_namespace(
            channels.GroupDMChannel, init_=False, make_icon_url=mock.Mock(return_value="icon-url-here.com")
        )()
        assert channel.icon_url == "icon-url-here.com"
        channel.make_icon_url.assert_called_once()

    def test_make_icon_url(self, model):
        assert model.make_icon_url(file_format="JPEG", size=16) == files.URL(
            "https://cdn.discordapp.com/channel-icons/136134/1a2b3c.jpeg?size=16"
        )

    def test_make_icon_url_without_optional_params(self, model):
        assert model.make_icon_url() == files.URL(
            "https://cdn.discordapp.com/channel-icons/136134/1a2b3c.png?size=4096"
        )

    def test_make_icon_url_when_hash_is_None(self, model):
        model.icon_hash = None
        assert model.make_icon_url() is None


class TestTextChannel:
    @pytest.fixture
    def model(self, mock_app):
        return hikari_test_helpers.mock_class_namespace(channels.TextableChannel)(
            app=mock_app, id=snowflakes.Snowflake(12345679), name="foo1", type=channels.ChannelType.GUILD_TEXT
        )

    @pytest.mark.asyncio
    async def test_fetch_history(self, model):
        model.app.rest.fetch_messages = mock.AsyncMock()

        await model.fetch_history(
            before=datetime.datetime(2020, 4, 1, 1, 0, 0),
            after=datetime.datetime(2020, 4, 1, 0, 0, 0),
            around=datetime.datetime(2020, 4, 1, 0, 30, 0),
        )

        model.app.rest.fetch_messages.assert_awaited_once_with(
            12345679,
            before=datetime.datetime(2020, 4, 1, 1, 0, 0),
            after=datetime.datetime(2020, 4, 1, 0, 0, 0),
            around=datetime.datetime(2020, 4, 1, 0, 30, 0),
        )

    @pytest.mark.asyncio
    async def test_fetch_message(self, model):
        model.app.rest.fetch_message = mock.AsyncMock()

        assert await model.fetch_message(133742069) is model.app.rest.fetch_message.return_value

        model.app.rest.fetch_message.assert_awaited_once_with(12345679, 133742069)

    @pytest.mark.asyncio
    async def test_fetch_pins(self, model):
        model.app.rest.fetch_pins = mock.AsyncMock()

        await model.fetch_pins()

        model.app.rest.fetch_pins.assert_awaited_once_with(12345679)

    @pytest.mark.asyncio
    async def test_pin_message(self, model):
        model.app.rest.pin_message = mock.AsyncMock()

        assert await model.pin_message(77790) is model.app.rest.pin_message.return_value

        model.app.rest.pin_message.assert_awaited_once_with(12345679, 77790)

    @pytest.mark.asyncio
    async def test_unpin_message(self, model):
        model.app.rest.unpin_message = mock.AsyncMock()

        assert await model.unpin_message(77790) is model.app.rest.unpin_message.return_value

        model.app.rest.unpin_message.assert_awaited_once_with(12345679, 77790)

    @pytest.mark.asyncio
    async def test_delete_messages(self, model):
        model.app.rest.delete_messages = mock.AsyncMock()

        await model.delete_messages([77790, 88890, 1800], 1337)

        model.app.rest.delete_messages.assert_awaited_once_with(12345679, [77790, 88890, 1800], 1337)

    @pytest.mark.asyncio
    async def test_send(self, model):
        model.app.rest.create_message = mock.AsyncMock()
        mock_attachment = object()
        mock_component = object()
        mock_components = [object(), object()]
        mock_embed = object()
        mock_embeds = object()
        mock_attachments = [object(), object(), object()]
        mock_reply = object()

        await model.send(
            content="test content",
            tts=True,
            attachment=mock_attachment,
            attachments=mock_attachments,
            component=mock_component,
            components=mock_components,
            embed=mock_embed,
            embeds=mock_embeds,
            sticker=543,
            stickers=[132, 65423],
            reply=mock_reply,
            reply_must_exist=False,
            mentions_everyone=False,
            user_mentions=[123, 456],
            role_mentions=[789, 567],
            mentions_reply=True,
            flags=6969,
        )

        model.app.rest.create_message.assert_awaited_once_with(
            channel=12345679,
            content="test content",
            tts=True,
            attachment=mock_attachment,
            attachments=mock_attachments,
            component=mock_component,
            components=mock_components,
            embed=mock_embed,
            embeds=mock_embeds,
            sticker=543,
            stickers=[132, 65423],
            reply=mock_reply,
            reply_must_exist=False,
            mentions_everyone=False,
            user_mentions=[123, 456],
            role_mentions=[789, 567],
            mentions_reply=True,
            flags=6969,
        )

    def test_trigger_typing(self, model):
        model.app.rest.trigger_typing = mock.Mock()

        model.trigger_typing()

        model.app.rest.trigger_typing.assert_called_once_with(12345679)


class TestGuildChannel:
    @pytest.fixture
    def model(self, mock_app):
        return hikari_test_helpers.mock_class_namespace(channels.GuildChannel)(
            app=mock_app,
            id=snowflakes.Snowflake(69420),
            name="foo1",
            type=channels.ChannelType.GUILD_VOICE,
            guild_id=snowflakes.Snowflake(123456789),
            parent_id=None,
        )

    def test_shard_id_property_when_not_shard_aware(self, model):
        model.app = None

        assert model.shard_id is None

    def test_shard_id_property_when_guild_id_is_not_None(self, model):
        model.app.shard_count = 3
        assert model.shard_id == 2

    @pytest.mark.asyncio
    async def test_fetch_guild(self, model):
        model.app.rest.fetch_guild = mock.AsyncMock()

        assert await model.fetch_guild() is model.app.rest.fetch_guild.return_value

        model.app.rest.fetch_guild.assert_awaited_once_with(123456789)

    @pytest.mark.asyncio
    async def test_edit(self, model):
        model.app.rest.edit_channel = mock.AsyncMock()

        result = await model.edit(
            name="Supa fast boike",
            bitrate=420,
            reason="left right",
            default_auto_archive_duration=123312,
            position=4423,
            topic="hi",
            nsfw=True,
            video_quality_mode=channels.VideoQualityMode.AUTO,
            user_limit=123321,
            rate_limit_per_user=54123123,
            region="us-west",
            parent_category=341123123123,
            permission_overwrites={123: "123"},
            flags=12,
            archived=True,
            auto_archive_duration=1234,
            locked=True,
            invitable=True,
            applied_tags=[12345, 54321],
        )

        assert result is model.app.rest.edit_channel.return_value
        model.app.rest.edit_channel.assert_awaited_once_with(
            69420,
            name="Supa fast boike",
            position=4423,
            topic="hi",
            nsfw=True,
            bitrate=420,
            video_quality_mode=channels.VideoQualityMode.AUTO,
            user_limit=123321,
            rate_limit_per_user=54123123,
            region="us-west",
            permission_overwrites={123: "123"},
            parent_category=341123123123,
            default_auto_archive_duration=123312,
            flags=12,
            archived=True,
            auto_archive_duration=1234,
            locked=True,
            invitable=True,
            applied_tags=[12345, 54321],
            reason="left right",
        )


class TestPermissibleGuildChannel:
    @pytest.fixture
    def model(self, mock_app):
        return hikari_test_helpers.mock_class_namespace(channels.PermissibleGuildChannel)(
            app=mock_app,
            id=snowflakes.Snowflake(69420),
            name="foo1",
            type=channels.ChannelType.GUILD_VOICE,
            guild_id=snowflakes.Snowflake(123456789),
            is_nsfw=True,
            parent_id=None,
            position=54,
            permission_overwrites=[],
        )

    @pytest.mark.asyncio
    async def test_edit_overwrite(self, model):
        model.app.rest.edit_permission_overwrite = mock.AsyncMock()
        user = mock.Mock(users.PartialUser)
        await model.edit_overwrite(
            333,
            target_type=user,
            allow=permissions.Permissions.BAN_MEMBERS,
            deny=permissions.Permissions.CONNECT,
            reason="vrooom vroom",
        )

        model.app.rest.edit_permission_overwrite.assert_called_once_with(
            69420,
            333,
            target_type=user,
            allow=permissions.Permissions.BAN_MEMBERS,
            deny=permissions.Permissions.CONNECT,
            reason="vrooom vroom",
        )

    @pytest.mark.asyncio
    async def test_edit_overwrite_target_type_none(self, model):
        model.app.rest.edit_permission_overwrite = mock.AsyncMock()
        user = mock.Mock(users.PartialUser)
        await model.edit_overwrite(
            user, allow=permissions.Permissions.BAN_MEMBERS, deny=permissions.Permissions.CONNECT, reason="vrooom vroom"
        )

        model.app.rest.edit_permission_overwrite.assert_called_once_with(
            69420,
            user,
            allow=permissions.Permissions.BAN_MEMBERS,
            deny=permissions.Permissions.CONNECT,
            reason="vrooom vroom",
        )

    @pytest.mark.asyncio
    async def test_remove_overwrite(self, model):
        model.app.rest.delete_permission_overwrite = mock.AsyncMock()

        await model.remove_overwrite(333)

        model.app.rest.delete_permission_overwrite.assert_called_once_with(69420, 333)

    def test_get_guild(self, model):
        guild = mock.Mock(id=123456789)
        model.app.cache.get_guild.side_effect = [guild]

        assert model.get_guild() == guild

        model.app.cache.get_guild.assert_called_once_with(123456789)

    def test_get_guild_when_guild_not_in_cache(self, model):
        model.app.cache.get_guild.side_effect = [None]

        assert model.get_guild() is None

        model.app.cache.get_guild.assert_called_once_with(123456789)

    def test_get_guild_when_no_cache_trait(self, model):
        model.app = object()

        assert model.get_guild() is None

    @pytest.mark.asyncio
    async def test_fetch_guild(self, model):
        model.app.rest.fetch_guild = mock.AsyncMock()

        assert await model.fetch_guild() == model.app.rest.fetch_guild.return_value

        model.app.rest.fetch_guild.assert_awaited_once_with(123456789)


class TestForumTag:
    @pytest.mark.parametrize(
        ("emoji", "expected_unicode_emoji", "expected_emoji_id"),
        [(123, None, 123), ("emoji", "emoji", None), (None, None, None)],
    )
    def test_emoji_parameters(self, emoji, expected_emoji_id, expected_unicode_emoji):
        tag = channels.ForumTag(name="testing", emoji=emoji)

        assert tag.emoji_id == expected_emoji_id
        assert tag.unicode_emoji == expected_unicode_emoji



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import contextlib
import re
import threading

import aiohttp
import aiohttp.abc
import aiohttp.web
import aiohttp.web_runner
import mock
import multidict

from hikari import applications
from hikari import files

try:
    import nacl.exceptions
    import nacl.signing

    nacl_present = True
except ModuleNotFoundError:
    nacl_present = False

import pytest

from hikari import errors
from hikari.impl import entity_factory as entity_factory_impl
from hikari.impl import interaction_server as interaction_server_impl
from hikari.impl import rest as rest_impl
from hikari.interactions import base_interactions
from tests.hikari import hikari_test_helpers


class MockWriter(aiohttp.abc.AbstractStreamWriter):
    def __init__(self):
        self.payload = bytearray()

    async def write(self, chunk: bytes) -> None:
        self.payload.extend(chunk)

    async def write_eof(self, chunk: bytes = b"") -> None:
        pass

    async def drain(self) -> None:
        pass

    def enable_compression(self, encoding: str = "deflate") -> None:
        pass

    def enable_chunking(self) -> None:
        pass

    async def write_headers(self, status_line: str, headers: "multidict.CIMultiDict[str]") -> None:
        pass


@pytest.mark.asyncio
class TestConsumeGeneratorListener:
    async def test_normal_behaviour(self):
        async def mock_generator_listener():
            nonlocal g_continued

            yield

            g_continued = True

        g_continued = False
        generator = mock_generator_listener()
        # The function expects the generator to have already yielded once
        await generator.__anext__()

        await interaction_server_impl._consume_generator_listener(generator)

        assert g_continued is True

    async def test_when_more_than_one_yield(self):
        async def mock_generator_listener():
            nonlocal g_continued

            yield

            g_continued = True

            yield

        g_continued = False
        generator = mock_generator_listener()
        # The function expects the generator to have already yielded once
        await generator.__anext__()

        loop = mock.Mock()
        with mock.patch.object(asyncio, "get_running_loop", return_value=loop):
            await interaction_server_impl._consume_generator_listener(generator)

        assert g_continued is True
        args, _ = loop.call_exception_handler.call_args_list[0]
        exception = args[0]["exception"]
        assert isinstance(exception, RuntimeError)
        assert exception.args == ("Generator listener yielded more than once, expected only one yield",)

    async def test_when_exception(self):
        async def mock_generator_listener():
            nonlocal g_continued, exception

            yield

            g_continued = True

            raise exception

        g_continued = False
        exception = ValueError("Some random exception")
        generator = mock_generator_listener()
        # The function expects the generator to have already yielded once
        await generator.__anext__()

        loop = mock.Mock()
        with mock.patch.object(asyncio, "get_running_loop", return_value=loop):
            await interaction_server_impl._consume_generator_listener(generator)

        assert g_continued is True
        args, _ = loop.call_exception_handler.call_args_list[0]
        assert args[0]["exception"] is exception


@pytest.fixture
def valid_edd25519():
    body = (
        b'{"application_id":"658822586720976907","channel_id":"938391701561679903","data":{"id":"870616445036421159","'
        b'name":"ping","type":1},"guild_id":"561884984214814744","guild_locale":"en-GB","id":"938421734825140264","loc'
        b'ale":"en-GB","member":{"avatar":null,"communication_disabled_until":null,"deaf":false,"is_pending":false,"jo'
        b'ined_at":"2019-03-31T12:10:19.616000+00:00","mute":false,"nick":"Snab","pending":false,"permissions":"219902'
        b'3255551","premium_since":null,"roles":["561885635074457600"],"user":{"avatar":"2549318b6a5f514a6c4a4379ed89a'
        b'469","discriminator":"6127","id":"115590097100865541","public_flags":131072,"username":"Faster Speeding"}},"'
        b'token":"aW50ZXJhY3Rpb246OTM4NDIxNzM0ODI1MTQwMjY0Ok1pR0t6dGt3T1Q4SkhHMnREQmJ2RXI4Vk5vaXZ0UzVMRTBqdVRLcmhnd1dY'
        b"dEd6d2dlTUZGMlNQRkRybGZJaHVuWHZva2lKaWQzcjh1ZEt5NzJtVTFKNzdGOVREOWtoNE5BYnlCdGlGaEZDMDVMY3VhbkF1a0ZZMnhGeU9q"
        b'OHY4","type":2,"version":1}'
    )
    signature = (
        b"\x8c*\xb2\x9e\x05\x0cSgc\xc9}A\xbd\x02.'-[\xb0\xa9\xbegN\xd5Z\x12\xa6 \xc5\x0b!\xd8c"
        b"B\x99\xf5W\x80\x07\xf2\x97\xba\x97\xcc\x17 L\xc53kG\xa0\x1c\x11\x8e|X\x05P\x81@.\xb5\x04"
    )
    timestamp = b"1643807576"
    return (body, signature, timestamp)


@pytest.fixture
def valid_payload():
    return {
        "application_id": "658822586720976907",
        "channel_id": "938391701561679903",
        "data": {"id": "870616445036421159", "name": "ping", "type": 1},
        "guild_id": "561884984214814744",
        "guild_locale": "en-GB",
        "id": "938421734825140264",
        "locale": "en-GB",
        "member": {
            "avatar": None,
            "communication_disabled_until": None,
            "deaf": False,
            "is_pending": False,
            "joined_at": "2019-03-31T12:10:19.616000+00:00",
            "mute": False,
            "nick": "Snab",
            "pending": False,
            "permissions": "2199023255551",
            "premium_since": None,
            "roles": ["561885635074457600"],
            "user": {
                "avatar": "2549318b6a5f514a6c4a4379ed89a469",
                "discriminator": "6127",
                "id": "115590097100865541",
                "public_flags": 131072,
                "username": "Faster Speeding",
            },
        },
        "token": (
            "aW50ZXJhY3Rpb246OTM4NDIxNzM0ODI1MTQwMjY0Ok1pR0t6dGt3T1Q4SkhHMnREQmJ2RXI"
            "4Vk5vaXZ0UzVMRTBqdVRLcmhnd1dYdEd6d2dlTUZGMlNQRkRybGZJaHVuWHZva2lKaWQzc"
            "jh1ZEt5NzJtVTFKNzdGOVREOWtoNE5BYnlCdGlGaEZDMDVMY3VhbkF1a0ZZMnhGeU9qOHY4"
        ),
        "type": 2,
        "version": 1,
    }


@pytest.fixture
def invalid_ed25519():
    body = (
        b'{"application_id":"658822586720976907","id":"838085779104202754","token":"aW50ZXJhY3Rpb246ODM4MDg1Nzc5MTA0MjA'
        b"yNzU0OmNhSk9QUU4wa1BKV21nTjFvSGhIbUp0QnQ1NjNGZFRtMlJVRlNjR0ttaDhtUGJrWUNvcmxYZnd2NTRLeUQ2c0hGS1YzTU03dFJ0V0s5"
        b'RWRBY0ltZTRTS0NneFFSYW1BbDZxSkpnMkEwejlkTldXZUh2OGwzbnBrMzhscURIMXUz","type":1,"user":{"avatar":"b333580bd947'
        b'4630226ff7b0a9696231","discriminator":"6127","id":"115590097100865541","public_flags":13'
        b'1072,"username":"Faster Speeding"},"version":1}'
    )
    signature = (
        b"\x0c4\xda!\xd9\xd5\x08<{a\x0c\xfa\xe6\xd2\x9e\xb3\xe0\x17r\x83\xa8\xb5\xda\xaa\x97\n\xb5\xe1\x92A|\x94\xbb"
        b"\x8aGu\xdb\xd6\x19\xd5\x94\x98\x08\xb4\x1a\xfaF@\xbbx\xc9\xa3\x8f\x1f\x13\t\xd81\xa3:\xa9%p\x0c"
    )
    timestamp = b"1619885620"
    return (body, signature, timestamp)


@pytest.fixture
def public_key():
    return b"\x12-\xdfX\xa8\x95\xd7\xe1\xb7o\xf5\xd0q\xb0\xaa\xc9\xb7v^*\xb5\x15\xe1\x1b\x7f\xca\xf9d\xdbT\x90\xc6"


@pytest.mark.skipif(nacl_present, reason="PyNacl is present")
def test_interaction_server_init_when_no_pynacl():
    with pytest.raises(
        RuntimeError,
        match=re.escape(
            "You must install the optional `hikari[server]` dependencies to use the default interaction server."
        ),
    ):
        interaction_server_impl.InteractionServer(entity_factory=mock.Mock(), rest_client=mock.Mock())


@pytest.mark.skipif(not nacl_present, reason="PyNacl not present")
class TestInteractionServer:
    @pytest.fixture
    def mock_entity_factory(self):
        return mock.Mock(entity_factory_impl.EntityFactoryImpl)

    @pytest.fixture
    def mock_rest_client(self):
        return mock.Mock(rest_impl.RESTClientImpl)

    @pytest.fixture
    def mock_interaction_server(
        self, mock_entity_factory: entity_factory_impl.EntityFactoryImpl, mock_rest_client: rest_impl.RESTClientImpl
    ):
        cls = hikari_test_helpers.mock_class_namespace(interaction_server_impl.InteractionServer, slots_=False)
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(rest_impl, "RESTClientImpl", return_value=mock_rest_client))

        with stack:
            return cls(entity_factory=mock_entity_factory, rest_client=mock_rest_client)

    def test___init__(
        self, mock_rest_client: rest_impl.RESTClientImpl, mock_entity_factory: entity_factory_impl.EntityFactoryImpl
    ):
        mock_dumps = object()
        mock_loads = object()

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(aiohttp.web, "Application"))

        with stack:
            result = interaction_server_impl.InteractionServer(
                dumps=mock_dumps,
                entity_factory=mock_entity_factory,
                loads=mock_loads,
                rest_client=mock_rest_client,
                public_key=None,
            )

        assert result._dumps is mock_dumps
        assert result._entity_factory is mock_entity_factory
        assert result._loads is mock_loads
        assert result._rest_client is mock_rest_client
        assert result._public_key is None

    def test___init___with_public_key(
        self, mock_rest_client: rest_impl.RESTClientImpl, mock_entity_factory: entity_factory_impl.EntityFactoryImpl
    ):
        mock_dumps = object()
        mock_loads = object()

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(aiohttp.web, "Application"))

        with stack:
            result = interaction_server_impl.InteractionServer(
                dumps=mock_dumps,
                entity_factory=mock_entity_factory,
                loads=mock_loads,
                rest_client=mock_rest_client,
                public_key=None,
            )

        assert result._public_key is None

    def test_is_alive_property_when_inactive(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        assert mock_interaction_server.is_alive is False

    def test_is_alive_property_when_active(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_interaction_server._server = object()

        assert mock_interaction_server.is_alive is True

    @pytest.mark.asyncio
    async def test___fetch_public_key_when_lock_is_None_gets_new_lock_and_doesnt_overwrite_existing_ones(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_rest_client: rest_impl.RESTClientImpl,
    ):
        mock_rest_client.token_type = "Bot"
        mock_interaction_server._application_fetch_lock = None
        mock_rest_client.fetch_application.return_value.public_key = (
            b"e\xb9\xf8\xac]eH\xb1\xe1D\xafaW\xdd\x1c.\xc1s\xfd<\x82\t\xeaO\xd4w\xaf\xc4\x1b\xd0\x8f\xc5"
        )
        results = []

        with mock.patch.object(asyncio, "Lock") as lock_class:
            # Run some times to make sure it does not overwrite it
            for _ in range(5):
                results.append(await mock_interaction_server._fetch_public_key())

        assert results[0] == nacl.signing.VerifyKey(mock_rest_client.fetch_application.return_value.public_key)
        assert all(result is results[0] for result in results)
        assert mock_interaction_server._application_fetch_lock is lock_class.return_value
        lock_class.assert_called_once_with()
        lock_class.return_value.__aenter__.assert_has_awaits([mock.call() for _ in range(5)])
        lock_class.return_value.__aexit__.assert_has_awaits([mock.call(None, None, None) for _ in range(5)])

    @pytest.mark.asyncio
    async def test__fetch_public_key_with_bearer_token(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_rest_client: rest_impl.RESTClientImpl,
    ):
        mock_rest_client.token_type = "Bearer"
        mock_interaction_server._application_fetch_lock = mock.AsyncMock()
        mock_rest_client.fetch_authorization.return_value.application.public_key = (
            b'\x81\xa9\xc0\xee"\xf0%\xd1CF\x82Uh\x16.>\x9b\xcf[\x1f\xa4\xfcsb\xc3\xf4x\xf9\xe0z\xad\xed'
        )

        result = await mock_interaction_server._fetch_public_key()

        assert result == nacl.signing.VerifyKey(
            mock_rest_client.fetch_authorization.return_value.application.public_key
        )
        mock_rest_client.fetch_authorization.assert_awaited_once()
        mock_rest_client.fetch_application.assert_not_called()
        mock_interaction_server._application_fetch_lock.__aenter__.assert_awaited_once()
        mock_interaction_server._application_fetch_lock.__aexit__.assert_awaited_once()

    @pytest.mark.asyncio
    async def test__fetch_public_key_fetch_with_bot_token(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_rest_client: rest_impl.RESTClientImpl,
    ):
        mock_rest_client.token_type = "Bot"
        mock_interaction_server._application_fetch_lock = mock.AsyncMock()
        mock_rest_client.fetch_application.return_value.public_key = (
            b"\xf3\xfd\xfc\x81\xfcU\x00\xe5;V\x15\xc6H\xab4Ip\x07\x1bR\xc2b9\x86\xa9\\e\xfa\xcbw\xd7\x0b"
        )

        result = await mock_interaction_server._fetch_public_key()

        assert result == nacl.signing.VerifyKey(mock_rest_client.fetch_application.return_value.public_key)
        mock_rest_client.fetch_authorization.assert_not_called()
        mock_rest_client.fetch_application.assert_awaited_once()
        mock_interaction_server._application_fetch_lock.__aenter__.assert_awaited_once()
        mock_interaction_server._application_fetch_lock.__aexit__.assert_awaited_once()

    @pytest.mark.asyncio
    async def test__fetch_public_key_when_public_key_already_set(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_lock = mock.AsyncMock()
        mock_public_key = object()
        mock_interaction_server._application_fetch_lock = mock_lock
        mock_interaction_server._public_key = mock_public_key

        assert await mock_interaction_server._fetch_public_key() is mock_public_key

        mock_lock.__aenter__.assert_awaited_once()
        mock_lock.__aexit__.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_aiohttp_hook(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_interaction_server.on_interaction = mock.AsyncMock(
            return_value=mock.Mock(
                payload=b"abody",
                files=[],
                status_code=200,
                headers={"header1": "ok"},
                content_type="ooga/booga",
                charset=None,
            )
        )
        request = mock.Mock(
            aiohttp.web.Request,
            content_type="application/json",
            headers={"X-Signature-Ed25519": "74726f656b70657769656f6a6b736939", "X-Signature-Timestamp": "123123"},
            read=mock.AsyncMock(return_value=b"bfddasdasd"),
        )

        result = await mock_interaction_server.aiohttp_hook(request)

        mock_interaction_server.on_interaction.assert_awaited_once_with(
            body=b"bfddasdasd", signature=b"troekpewieojksi9", timestamp=b"123123"
        )
        assert result.body == b"abody"
        assert result.content_type == "ooga/booga"
        assert result.headers == {"header1": "ok", "Content-Type": "ooga/booga"}
        assert result.status == 200

    @pytest.mark.asyncio
    async def test_aiohttp_hook_when_no_other_headers(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_interaction_server.on_interaction = mock.AsyncMock(
            return_value=mock.Mock(
                payload=b"abody", files=[], headers=None, status_code=200, content_type="ooga/booga", charset=None
            )
        )
        request = mock.Mock(
            aiohttp.web.Request,
            content_type="application/json",
            headers={"X-Signature-Ed25519": "74726f656b70657769656f6a6b736939", "X-Signature-Timestamp": "123123"},
            read=mock.AsyncMock(return_value=b"bfddasdasd"),
        )

        result = await mock_interaction_server.aiohttp_hook(request)

        mock_interaction_server.on_interaction.assert_awaited_once_with(
            body=b"bfddasdasd", signature=b"troekpewieojksi9", timestamp=b"123123"
        )
        assert result.body == b"abody"
        assert result.content_type == "ooga/booga"
        assert result.headers == {"Content-Type": "ooga/booga"}
        assert result.status == 200

    @pytest.mark.asyncio
    async def test_aiohttp_hook_when_files(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_interaction_server.on_interaction = mock.AsyncMock(
            return_value=mock.Mock(
                payload=b"abody",
                files=[files.Bytes("x" * 329, "meow.txt"), files.Bytes("y" * 124, "nyaa.txt")],
                status_code=200,
                headers={"header1": "ok"},
                content_type="ooga/booga",
            )
        )
        request = mock.Mock(
            aiohttp.web.Request,
            content_type="application/json",
            headers={"X-Signature-Ed25519": "74726f656b70657769656f6a6b736939", "X-Signature-Timestamp": "123123"},
            read=mock.AsyncMock(return_value=b"bfddasdasd"),
        )

        result = await mock_interaction_server.aiohttp_hook(request)

        mock_interaction_server.on_interaction.assert_awaited_once_with(
            body=b"bfddasdasd", signature=b"troekpewieojksi9", timestamp=b"123123"
        )
        assert isinstance(result.body, aiohttp.MultipartWriter)
        assert result.content_type == "multipart/form-data"
        assert result.headers == {
            "header1": "ok",
            "Content-Type": f"multipart/form-data; boundary={result.body.boundary}",
        }
        assert result.status == 200

        mock_writer = MockWriter()
        await result.body.write(mock_writer)

        boundary = result.body.boundary.encode()
        assert mock_writer.payload == bytearray(
            b"--" + boundary + b"""\r\nContent-Type: ooga/booga\r\nContent-Disposition: form-data; name="payload_json"""
            b""""\r\n\r\nabody\r\n--""" + boundary + b"""\r\nContent-Type: text/plain\r\nContent-Disposition: """
            b"""form-data; name="files[0]"; filename="meow.txt"\r\n\r\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"""
            b"""xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"""
            b"""xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"""
            b"""xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"""
            b"""\r\n--""" + boundary + b"""\r\nContent-Type: text/plain\r\nContent-Disposition: form-data; name="fil"""
            b"""es[1]"; filename="nyaa.txt"\r\n\r\nyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy"""
            b"""yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy\r\n--""" + boundary + b"""--\r\n"""
        )

    @pytest.mark.asyncio
    async def test_aiohttp_hook_for_unsupported_media_type(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        request = mock.Mock(aiohttp.web.Request, content_type="notjson")

        result = await mock_interaction_server.aiohttp_hook(request)

        assert result.body == b"Unsupported Media Type"
        assert result.charset == "UTF-8"
        assert result.content_type == "text/plain"
        assert result.status == 415

    @pytest.mark.asyncio
    async def test_aiohttp_hook_with_missing_ed25519_header(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        request = mock.Mock(aiohttp.web.Request, content_type="application/json", headers={"X-Signature-Timestamp": ""})

        result = await mock_interaction_server.aiohttp_hook(request)

        assert result.body == b"Missing or invalid required request signature header(s)"
        assert result.charset == "UTF-8"
        assert result.content_type == "text/plain"
        assert result.status == 400

    @pytest.mark.asyncio
    async def test_aiohttp_hook_with_missing_timestamp_header(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        request = mock.Mock(aiohttp.web.Request, content_type="application/json", headers={"X-Signature-Ed25519": ""})

        result = await mock_interaction_server.aiohttp_hook(request)

        assert result.body == b"Missing or invalid required request signature header(s)"
        assert result.charset == "UTF-8"
        assert result.content_type == "text/plain"
        assert result.status == 400

    @pytest.mark.asyncio
    async def test_aiohttp_hook_with_invalid_ed25519_header(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        request = mock.Mock(
            aiohttp.web.Request,
            content_type="application/json",
            headers={"X-Signature-Ed25519": "'#';[][];['345", "X-Signature-Timestamp": ""},
        )

        result = await mock_interaction_server.aiohttp_hook(request)

        assert result.body == b"Missing or invalid required request signature header(s)"
        assert result.charset == "UTF-8"
        assert result.content_type == "text/plain"
        assert result.status == 400

    @pytest.mark.asyncio
    async def test_aiohttp_hook_when_payload_too_large(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        request = mock.Mock(
            aiohttp.web.Request,
            content_type="application/json",
            headers={"X-Signature-Ed25519": "74726f656b70657769656f6a6b736939", "X-Signature-Timestamp": "123123"},
            read=mock.AsyncMock(side_effect=aiohttp.web.HTTPRequestEntityTooLarge(123, 321)),
        )

        result = await mock_interaction_server.aiohttp_hook(request)

        assert result.body == b"Payload too large"
        assert result.charset == "UTF-8"
        assert result.content_type == "text/plain"
        assert result.status == 413

    @pytest.mark.asyncio
    async def test_aiohttp_hook_when_no_body(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        request = mock.Mock(
            aiohttp.web.Request,
            content_type="application/json",
            headers={"X-Signature-Ed25519": "74726f656b70657769656f6a6b736939", "X-Signature-Timestamp": "123123"},
            read=mock.AsyncMock(return_value=""),
        )

        result = await mock_interaction_server.aiohttp_hook(request)

        assert result.body == b"POST request must have a body"
        assert result.charset == "UTF-8"
        assert result.content_type == "text/plain"
        assert result.status == 400

    @pytest.mark.asyncio
    async def test_close(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_runner = mock.AsyncMock()
        mock_event = mock.Mock()
        mock_interaction_server._is_closing = False
        mock_interaction_server._server = mock_runner
        mock_interaction_server._close_event = mock_event
        generator_listener_1 = mock.Mock()
        generator_listener_2 = mock.Mock()
        generator_listener_3 = mock.Mock()
        generator_listener_4 = mock.Mock()
        mock_interaction_server._running_generator_listeners = [
            generator_listener_1,
            generator_listener_2,
            generator_listener_3,
            generator_listener_4,
        ]

        with mock.patch.object(asyncio, "gather", new=mock.AsyncMock()) as gather:
            await mock_interaction_server.close()

        mock_runner.shutdown.assert_awaited_once()
        mock_runner.cleanup.assert_awaited_once()
        mock_event.set.assert_called_once()
        assert mock_interaction_server._is_closing is False
        assert mock_interaction_server._running_generator_listeners == set()
        gather.assert_awaited_once_with(
            generator_listener_1, generator_listener_2, generator_listener_3, generator_listener_4
        )

    @pytest.mark.asyncio
    async def test_close_when_closing(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_runner = mock.AsyncMock()
        mock_event = mock.Mock()
        mock_interaction_server._server = mock_runner
        mock_interaction_server._close_event = mock_event
        mock_interaction_server._is_closing = True
        mock_interaction_server.join = mock.AsyncMock()
        mock_listener = object()
        mock_interaction_server._running_generator_listeners = [mock_listener]

        await mock_interaction_server.close()

        mock_runner.shutdown.assert_not_called()
        mock_runner.cleanup.assert_not_called()
        mock_event.set.assert_not_called()
        mock_interaction_server.join.assert_awaited_once()
        assert mock_interaction_server._running_generator_listeners == [mock_listener]

    @pytest.mark.asyncio
    async def test_close_when_not_running(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        with pytest.raises(errors.ComponentStateConflictError):
            await mock_interaction_server.close()

    @pytest.mark.asyncio
    async def test_join(self, mock_interaction_server):
        mock_event = mock.AsyncMock()
        mock_interaction_server._server = object()
        mock_interaction_server._close_event = mock_event

        await mock_interaction_server.join()

        mock_event.wait.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_join_when_not_running(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        with pytest.raises(errors.ComponentStateConflictError):
            await mock_interaction_server.join()

    @pytest.mark.asyncio
    async def test_on_interaction(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
        public_key: bytes,
        valid_edd25519: bytes,
        valid_payload: bytes,
    ):
        mock_interaction_server._public_key = nacl.signing.VerifyKey(public_key)
        mock_file_1 = mock.Mock()
        mock_file_2 = mock.Mock()
        mock_entity_factory.deserialize_interaction.return_value = base_interactions.PartialInteraction(
            app=None,
            id=123,
            application_id=541324,
            type=2,
            token="ok",
            version=1,
            authorizing_integration_owners={},
            context=applications.ApplicationContextType.GUILD,
            app_permissions=123123123,
            user=mock.Mock(),
            member=mock.Mock(),
            channel=mock.Mock(),
            guild_id=123123,
            guild_locale="en-GB",
            locale="es-ES",
            entitlements=[],
        )
        mock_builder = mock.Mock(build=mock.Mock(return_value=({"ok": "No boomer"}, [mock_file_1, mock_file_2])))
        mock_listener = mock.AsyncMock(return_value=mock_builder)
        mock_interaction_server.set_listener(base_interactions.PartialInteraction, mock_listener)

        result = await mock_interaction_server.on_interaction(*valid_edd25519)

        mock_listener.assert_awaited_once_with(mock_entity_factory.deserialize_interaction.return_value)
        mock_builder.build.assert_called_once_with(mock_entity_factory)
        mock_entity_factory.deserialize_interaction.assert_called_once_with(valid_payload)
        assert result.content_type == "application/json"
        assert result.charset == "UTF-8"
        assert result.files == [mock_file_1, mock_file_2]
        assert result.headers is None
        assert result.payload == b'{"ok":"No boomer"}'
        assert result.status_code == 200

    @pytest.mark.asyncio
    async def test_on_interaction_with_generator_listener(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
        public_key: bytes,
        valid_edd25519: bytes,
        valid_payload: bytes,
    ):
        async def mock_generator_listener(event):
            nonlocal g_called, g_complete

            g_called = True
            assert event is mock_entity_factory.deserialize_interaction.return_value

            yield mock_builder

            g_complete = True

        mock_interaction_server._public_key = nacl.signing.VerifyKey(public_key)
        mock_file_1 = mock.Mock()
        mock_file_2 = mock.Mock()
        mock_entity_factory.deserialize_interaction.return_value = base_interactions.PartialInteraction(
            app=None,
            id=123,
            application_id=541324,
            type=2,
            token="ok",
            version=1,
            authorizing_integration_owners={},
            context=applications.ApplicationContextType.GUILD,
            app_permissions=123123123,
            user=mock.Mock(),
            member=mock.Mock(),
            channel=mock.Mock(),
            guild_id=123123,
            guild_locale="en-GB",
            locale="es-ES",
            entitlements=[],
        )
        mock_builder = mock.Mock(build=mock.Mock(return_value=({"ok": "No boomer"}, [mock_file_1, mock_file_2])))
        g_called = False
        g_complete = False
        mock_interaction_server.set_listener(base_interactions.PartialInteraction, mock_generator_listener)

        result = await mock_interaction_server.on_interaction(*valid_edd25519)

        mock_builder.build.assert_called_once_with(mock_entity_factory)
        mock_entity_factory.deserialize_interaction.assert_called_once_with(valid_payload)
        assert result.content_type == "application/json"
        assert result.charset == "UTF-8"
        assert result.files == [mock_file_1, mock_file_2]
        assert result.headers is None
        assert result.payload == b'{"ok":"No boomer"}'
        assert result.status_code == 200

        assert g_called is True
        assert g_complete is False
        assert len(mock_interaction_server._running_generator_listeners) != 0
        # Give some time for the task to complete
        await asyncio.sleep(0.01)

        assert g_complete is True
        assert len(mock_interaction_server._running_generator_listeners) == 0

    @pytest.mark.asyncio
    async def test_on_interaction_returns_no_content(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
        public_key: bytes,
        valid_edd25519: bytes,
        valid_payload: bytes,
    ):
        mock_interaction_server._public_key = nacl.signing.VerifyKey(public_key)
        mock_file_1 = mock.Mock()
        mock_file_2 = mock.Mock()
        mock_entity_factory.deserialize_interaction.return_value = base_interactions.PartialInteraction(
            app=None,
            id=123,
            application_id=541324,
            type=2,
            token="ok",
            version=1,
            authorizing_integration_owners={},
            context=applications.ApplicationContextType.GUILD,
            app_permissions=123123123,
            user=mock.Mock(),
            member=mock.Mock(),
            channel=mock.Mock(),
            guild_id=123123,
            guild_locale="en-GB",
            locale="es-ES",
            entitlements=[],
        )

        mock_listener = mock.AsyncMock(return_value=None)
        mock_interaction_server.set_listener(base_interactions.PartialInteraction, mock_listener)

        result = await mock_interaction_server.on_interaction(*valid_edd25519)
        assert result.payload is None
        assert len(result.files) == 0
        assert result.status_code == 204

    @pytest.mark.asyncio
    async def test_on_interaction_calls__fetch_public_key(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_fetcher = mock.AsyncMock(
            return_value=mock.Mock(verify=mock.Mock(side_effect=nacl.exceptions.BadSignatureError))
        )
        mock_interaction_server._public_key = None
        mock_interaction_server._fetch_public_key = mock_fetcher

        result = await mock_interaction_server.on_interaction(b"body", b"signature", b"timestamp")

        mock_fetcher.assert_awaited_once()
        mock_fetcher.return_value.verify.assert_called_once_with(b"timestampbody", b"signature")
        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Invalid request signature"
        assert result.status_code == 400

    @pytest.mark.asyncio
    async def test_on_interaction_when_public_key_mismatch(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        public_key: bytes,
        invalid_ed25519: bytes,
    ):
        mock_interaction_server._public_key = nacl.signing.VerifyKey(public_key)

        result = await mock_interaction_server.on_interaction(*invalid_ed25519)

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Invalid request signature"
        assert result.status_code == 400

    @pytest.mark.parametrize("body", [b"not a json", b"\x80abc"])
    @pytest.mark.asyncio
    async def test_on_interaction_when_bad_body(
        self, mock_interaction_server: interaction_server_impl.InteractionServer, body: bytes
    ):
        mock_interaction_server._public_key = mock.Mock()

        result = await mock_interaction_server.on_interaction(body, b"signature", b"timestamp")

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Invalid JSON body"
        assert result.status_code == 400

    @pytest.mark.asyncio
    async def test_on_interaction_when_missing_type_key(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_interaction_server._public_key = mock.Mock()

        result = await mock_interaction_server.on_interaction(b'{"key": "OK"}', b"signature", b"timestamp")

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Missing required 'type' field in payload"
        assert result.status_code == 400

    @pytest.mark.asyncio
    async def test_on_interaction_on_ping(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_interaction_server._public_key = mock.Mock()

        result = await mock_interaction_server.on_interaction(b'{"type":1}', b"signature", b"timestamp")

        assert result.content_type == "application/json"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b'{"type":1}'
        assert result.status_code == 200

    @pytest.mark.asyncio
    async def test_on_interaction_on_deserialize_unrecognised_entity_error(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
    ):
        mock_interaction_server._public_key = mock.Mock()
        mock_entity_factory.deserialize_interaction.side_effect = errors.UnrecognisedEntityError("blah")

        result = await mock_interaction_server.on_interaction(b'{"type": 2}', b"signature", b"timestamp")

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Interaction type not implemented"
        assert result.status_code == 501

    @pytest.mark.asyncio
    async def test_on_interaction_on_failed_deserialize(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
    ):
        mock_interaction_server._public_key = mock.Mock()
        mock_exception = TypeError("OK")
        mock_entity_factory.deserialize_interaction.side_effect = mock_exception

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            result = await mock_interaction_server.on_interaction(b'{"type": 2}', b"signature", b"timestamp")

            get_running_loop.return_value.call_exception_handler.assert_called_once_with(
                {
                    "message": "Exception occurred during interaction deserialization",
                    "payload": {"type": 2},
                    "exception": mock_exception,
                }
            )

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Exception occurred during interaction deserialization"
        assert result.status_code == 500

    @pytest.mark.asyncio
    async def test_on_interaction_on_dispatch_error(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
    ):
        mock_interaction_server._public_key = mock.Mock()
        mock_exception = TypeError("OK")
        mock_entity_factory.deserialize_interaction.return_value = base_interactions.PartialInteraction(
            app=None,
            id=123,
            application_id=541324,
            type=2,
            token="ok",
            version=1,
            authorizing_integration_owners={},
            context=applications.ApplicationContextType.GUILD,
            app_permissions=123123123,
            user=mock.Mock(),
            member=mock.Mock(),
            channel=mock.Mock(),
            guild_id=123123,
            guild_locale="en-GB",
            locale="es-ES",
            entitlements=[],
        )
        mock_interaction_server.set_listener(
            base_interactions.PartialInteraction, mock.Mock(side_effect=mock_exception)
        )

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            result = await mock_interaction_server.on_interaction(b'{"type": 2}', b"signature", b"timestamp")

            get_running_loop.return_value.call_exception_handler.assert_called_once_with(
                {"message": "Exception occurred during interaction dispatch", "exception": mock_exception}
            )

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Exception occurred during interaction dispatch"
        assert result.status_code == 500

    @pytest.mark.asyncio
    async def test_on_interaction_when_response_builder_error(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
    ):
        mock_interaction_server._public_key = mock.Mock()
        mock_exception = TypeError("OK")
        mock_entity_factory.deserialize_interaction.return_value = base_interactions.PartialInteraction(
            app=None,
            id=123,
            application_id=541324,
            type=2,
            token="ok",
            version=1,
            authorizing_integration_owners={},
            context=applications.ApplicationContextType.GUILD,
            app_permissions=123123123,
            user=mock.Mock(),
            member=mock.Mock(),
            channel=mock.Mock(),
            guild_id=123123,
            guild_locale="en-GB",
            locale="es-ES",
            entitlements=[],
        )
        mock_builder = mock.Mock(build=mock.Mock(side_effect=mock_exception))
        mock_interaction_server.set_listener(
            base_interactions.PartialInteraction, mock.AsyncMock(return_value=mock_builder)
        )

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            result = await mock_interaction_server.on_interaction(b'{"type": 2}', b"signature", b"timestamp")

            get_running_loop.return_value.call_exception_handler.assert_called_once_with(
                {"message": "Exception occurred during interaction dispatch", "exception": mock_exception}
            )

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Exception occurred during interaction dispatch"
        assert result.status_code == 500

    @pytest.mark.asyncio
    async def test_on_interaction_when_json_encode_fails(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
    ):
        mock_interaction_server._public_key = mock.Mock()
        mock_exception = TypeError("OK")
        mock_interaction_server._dumps = mock.Mock(side_effect=mock_exception)
        mock_entity_factory.deserialize_interaction.return_value = base_interactions.PartialInteraction(
            app=None,
            id=123,
            application_id=541324,
            type=2,
            token="ok",
            version=1,
            authorizing_integration_owners={},
            context=applications.ApplicationContextType.GUILD,
            app_permissions=123123123,
            user=mock.Mock(),
            member=mock.Mock(),
            channel=mock.Mock(),
            guild_id=123123,
            guild_locale="en-GB",
            locale="es-ES",
            entitlements=[],
        )
        mock_builder = mock.Mock(build=mock.Mock(return_value=({"ok": "No"}, [])))
        mock_interaction_server.set_listener(
            base_interactions.PartialInteraction, mock.AsyncMock(return_value=mock_builder)
        )

        with mock.patch.object(asyncio, "get_running_loop") as get_running_loop:
            result = await mock_interaction_server.on_interaction(b'{"type": 2}', b"signature", b"timestamp")

            get_running_loop.return_value.call_exception_handler.assert_called_once_with(
                {"message": "Exception occurred during interaction dispatch", "exception": mock_exception}
            )

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Exception occurred during interaction dispatch"
        assert result.status_code == 500

    @pytest.mark.asyncio
    async def test_on_interaction_when_no_registered_listener(
        self,
        mock_interaction_server: interaction_server_impl.InteractionServer,
        mock_entity_factory: entity_factory_impl.EntityFactoryImpl,
    ):
        mock_interaction_server._public_key = mock.Mock()

        result = await mock_interaction_server.on_interaction(b'{"type": 2}', b"signature", b"timestamp")

        assert result.content_type == "text/plain"
        assert result.charset == "UTF-8"
        assert result.files == ()
        assert result.headers is None
        assert result.payload == b"Handler not set for this interaction type"
        assert result.status_code == 501

    @pytest.mark.asyncio
    async def test_start(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_context = object()
        mock_socket = object()
        mock_interaction_server._is_closing = True
        mock_interaction_server._fetch_public_key = mock.AsyncMock()
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(aiohttp.web, "TCPSite", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web, "UnixSite", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web, "SockSite", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web_runner, "AppRunner", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web, "Application"))
        stack.enter_context(mock.patch.object(asyncio, "Event"))

        with stack:
            await mock_interaction_server.start(
                backlog=123123,
                host="hoototototo",
                port=123123123,
                path="hshshshshsh",
                reuse_address=True,
                reuse_port=False,
                socket=mock_socket,
                shutdown_timeout=3232.3232,
                ssl_context=mock_context,
            )

            mock_interaction_server._fetch_public_key.assert_awaited_once_with()

            aiohttp.web.Application.assert_called_once_with()
            aiohttp.web.Application.return_value.add_routes.assert_called_once_with(
                [aiohttp.web.post("/", mock_interaction_server.aiohttp_hook)]
            )
            aiohttp.web_runner.AppRunner.assert_called_once_with(
                aiohttp.web.Application.return_value, access_log=interaction_server_impl._LOGGER
            )
            aiohttp.web_runner.AppRunner.return_value.setup.assert_awaited_once()
            aiohttp.web.TCPSite.assert_called_once_with(
                aiohttp.web_runner.AppRunner.return_value,
                "hoototototo",
                port=123123123,
                shutdown_timeout=3232.3232,
                ssl_context=mock_context,
                backlog=123123,
                reuse_address=True,
                reuse_port=False,
            )
            aiohttp.web.UnixSite.assert_called_once_with(
                aiohttp.web_runner.AppRunner.return_value,
                "hshshshshsh",
                shutdown_timeout=3232.3232,
                ssl_context=mock_context,
                backlog=123123,
            )
            aiohttp.web.SockSite.assert_called_once_with(
                aiohttp.web_runner.AppRunner.return_value,
                mock_socket,
                shutdown_timeout=3232.3232,
                ssl_context=mock_context,
                backlog=123123,
            )
            aiohttp.web.TCPSite.return_value.start.assert_awaited_once()
            aiohttp.web.UnixSite.return_value.start.assert_awaited_once()
            aiohttp.web.SockSite.return_value.start.assert_awaited_once()
            assert mock_interaction_server._close_event is asyncio.Event.return_value
            assert mock_interaction_server._is_closing is False

    @pytest.mark.asyncio
    async def test_start_with_default_behaviour(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_context = object()
        mock_interaction_server._fetch_public_key = mock.AsyncMock()
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(aiohttp.web, "TCPSite", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web_runner, "AppRunner", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web, "Application"))

        with stack:
            await mock_interaction_server.start(ssl_context=mock_context)

            mock_interaction_server._fetch_public_key.assert_awaited_once_with()

            aiohttp.web.Application.assert_called_once_with()
            aiohttp.web.Application.return_value.add_routes.assert_called_once_with(
                [aiohttp.web.post("/", mock_interaction_server.aiohttp_hook)]
            )
            aiohttp.web_runner.AppRunner.assert_called_once_with(
                aiohttp.web.Application.return_value, access_log=interaction_server_impl._LOGGER
            )
            aiohttp.web_runner.AppRunner.return_value.setup.assert_awaited_once()
            aiohttp.web.TCPSite.assert_called_once_with(
                aiohttp.web_runner.AppRunner.return_value,
                port=None,
                shutdown_timeout=60.0,
                ssl_context=mock_context,
                backlog=128,
                reuse_address=None,
                reuse_port=None,
            )
            aiohttp.web.TCPSite.return_value.start.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_start_with_default_behaviour_and_not_main_thread(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_context = object()
        mock_interaction_server._fetch_public_key = mock.AsyncMock()
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(aiohttp.web, "TCPSite", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web_runner, "AppRunner", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web, "Application"))
        stack.enter_context(mock.patch.object(threading, "current_thread"))

        with stack:
            await mock_interaction_server.start(ssl_context=mock_context)

            mock_interaction_server._fetch_public_key.assert_awaited_once_with()

            aiohttp.web.Application.assert_called_once_with()
            aiohttp.web.Application.return_value.add_routes.assert_called_once_with(
                [aiohttp.web.post("/", mock_interaction_server.aiohttp_hook)]
            )
            aiohttp.web_runner.AppRunner.assert_called_once_with(
                aiohttp.web.Application.return_value, access_log=interaction_server_impl._LOGGER
            )
            aiohttp.web_runner.AppRunner.return_value.setup.assert_awaited_once()
            aiohttp.web.TCPSite.assert_called_once_with(
                aiohttp.web_runner.AppRunner.return_value,
                port=None,
                shutdown_timeout=60.0,
                ssl_context=mock_context,
                backlog=128,
                reuse_address=None,
                reuse_port=None,
            )
            aiohttp.web.TCPSite.return_value.start.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_start_with_multiple_hosts(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_context = object()
        mock_interaction_server._fetch_public_key = mock.AsyncMock()
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(aiohttp.web, "TCPSite", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web_runner, "AppRunner", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web, "Application"))

        with stack:
            await mock_interaction_server.start(ssl_context=mock_context, host=["123", "4312"], port=453123)

            mock_interaction_server._fetch_public_key.assert_awaited_once_with()

            aiohttp.web.Application.assert_called_once_with()
            aiohttp.web.Application.return_value.add_routes.assert_called_once_with(
                [aiohttp.web.post("/", mock_interaction_server.aiohttp_hook)]
            )
            aiohttp.web_runner.AppRunner.assert_called_once_with(
                aiohttp.web.Application.return_value, access_log=interaction_server_impl._LOGGER
            )
            aiohttp.web_runner.AppRunner.return_value.setup.assert_awaited_once()
            aiohttp.web.TCPSite.assert_has_calls(
                [
                    mock.call(
                        aiohttp.web_runner.AppRunner.return_value,
                        "123",
                        port=453123,
                        shutdown_timeout=60.0,
                        ssl_context=mock_context,
                        backlog=128,
                        reuse_address=None,
                        reuse_port=None,
                    ),
                    mock.call(
                        aiohttp.web_runner.AppRunner.return_value,
                        "4312",
                        port=453123,
                        shutdown_timeout=60.0,
                        ssl_context=mock_context,
                        backlog=128,
                        reuse_address=None,
                        reuse_port=None,
                    ),
                ]
            )
            aiohttp.web.TCPSite.return_value.start.assert_has_awaits([mock.call(), mock.call()])

    @pytest.mark.asyncio
    async def test_start_when_no_tcp_sites(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_socket = object()
        mock_context = object()
        mock_interaction_server._fetch_public_key = mock.AsyncMock()
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(aiohttp.web, "TCPSite", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web_runner, "AppRunner", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web, "Application"))
        stack.enter_context(mock.patch.object(aiohttp.web, "SockSite", return_value=mock.AsyncMock()))
        stack.enter_context(mock.patch.object(aiohttp.web, "UnixSite", return_value=mock.AsyncMock()))

        with stack:
            await mock_interaction_server.start(ssl_context=mock_context, socket=mock_socket)

            mock_interaction_server._fetch_public_key.assert_awaited_once_with()

            aiohttp.web.Application.assert_called_once_with()
            aiohttp.web.Application.return_value.add_routes.assert_called_once_with(
                [aiohttp.web.post("/", mock_interaction_server.aiohttp_hook)]
            )
            aiohttp.web_runner.AppRunner.assert_called_once_with(
                aiohttp.web.Application.return_value, access_log=interaction_server_impl._LOGGER
            )
            aiohttp.web_runner.AppRunner.return_value.setup.assert_awaited_once()
            aiohttp.web.TCPSite.assert_not_called()
            aiohttp.web.UnixSite.assert_not_called()
            aiohttp.web.SockSite.assert_called_once_with(
                aiohttp.web_runner.AppRunner.return_value,
                mock_socket,
                shutdown_timeout=60.0,
                ssl_context=mock_context,
                backlog=128,
            )
            aiohttp.web.SockSite.return_value.start.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_start_when_already_running(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_interaction_server._server = object()

        with pytest.raises(errors.ComponentStateConflictError):
            await mock_interaction_server.start()

    def test_get_listener_when_unknown(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        assert mock_interaction_server.get_listener(base_interactions.PartialInteraction) is None

    def test_get_listener_when_registered(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_listener = object()
        mock_interaction_server.set_listener(base_interactions.PartialInteraction, mock_listener)

        assert mock_interaction_server.get_listener(base_interactions.PartialInteraction) is mock_listener

    def test_set_listener(self, mock_interaction_server: interaction_server_impl.InteractionServer):
        mock_listener = object()

        mock_interaction_server.set_listener(base_interactions.PartialInteraction, mock_listener)

        assert mock_interaction_server.get_listener(base_interactions.PartialInteraction) is mock_listener

    def test_set_listener_when_already_registered_without_replace(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_interaction_server.set_listener(base_interactions.PartialInteraction, object())

        with pytest.raises(TypeError):
            mock_interaction_server.set_listener(base_interactions.PartialInteraction, object())

    def test_set_listener_when_already_registered_with_replace(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_listener = object()
        mock_interaction_server.set_listener(base_interactions.PartialInteraction, object())

        mock_interaction_server.set_listener(base_interactions.PartialInteraction, mock_listener, replace=True)

        assert mock_interaction_server.get_listener(base_interactions.PartialInteraction) is mock_listener

    def test_set_listener_when_removing_listener(
        self, mock_interaction_server: interaction_server_impl.InteractionServer
    ):
        mock_interaction_server.set_listener(base_interactions.PartialInteraction, object())
        mock_interaction_server.set_listener(base_interactions.PartialInteraction, None)

        assert mock_interaction_server.get_listener(base_interactions.PartialInteraction) is None



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Base classes and enums inherited and used throughout the interactions flow."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "DEFERRED_RESPONSE_TYPES",
    "MESSAGE_RESPONSE_TYPES",
    "DeferredResponseTypesT",
    "InteractionChannel",
    "InteractionMember",
    "InteractionType",
    "MessageResponseMixin",
    "MessageResponseTypesT",
    "ModalResponseMixin",
    "PartialInteraction",
    "PartialInteractionMetadata",
    "ResolvedOptionData",
    "ResponseType",
)

import typing

import attrs

from hikari import channels
from hikari import guilds
from hikari import snowflakes
from hikari import traits
from hikari import undefined
from hikari import webhooks
from hikari.internal import attrs_extensions
from hikari.internal import enums
from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    from hikari import applications
    from hikari import embeds as embeds_
    from hikari import files
    from hikari import locales
    from hikari import messages
    from hikari import monetization
    from hikari import permissions as permissions_
    from hikari import users
    from hikari.api import special_endpoints


_CommandResponseTypesT = typing.TypeVar("_CommandResponseTypesT", bound=int)


@typing.final
class InteractionType(int, enums.Enum):
    """The type of an interaction."""

    # PING isn't here as it should be handled as internal detail of the REST
    # server rather than as a part of the public interface.
    APPLICATION_COMMAND = 2
    """An interaction triggered by a user calling an application command."""

    MESSAGE_COMPONENT = 3
    """An interaction triggered by a user calling a message component."""

    AUTOCOMPLETE = 4
    """An interaction triggered by a user typing in a slash command option."""

    MODAL_SUBMIT = 5
    """An interaction triggered by a user submitting a modal."""


@typing.final
class ResponseType(int, enums.Enum):
    """The type of an interaction response."""

    # PONG isn't here as it should be handled as internal detail of the REST
    # server rather than as a part of the public interface.

    # Type 2 and 3 aren't included as they were deprecated/removed by Discord.
    MESSAGE_CREATE = 4
    """An immediate message response to an interaction.

    * [`hikari.interactions.base_interactions.InteractionType.APPLICATION_COMMAND`][]
    * [`hikari.interactions.base_interactions.InteractionType.MESSAGE_COMPONENT`][]
    """

    DEFERRED_MESSAGE_CREATE = 5
    """Acknowledge an interaction with the intention to edit in a message response later.

    The user will see a loading state when this type is used until this
    interaction expires or a message response is edited in over REST.

    This is valid for the following interaction types:

    * [`hikari.interactions.base_interactions.InteractionType.APPLICATION_COMMAND`][]
    * [`hikari.interactions.base_interactions.InteractionType.MESSAGE_COMPONENT`][]
    """

    DEFERRED_MESSAGE_UPDATE = 6
    """Acknowledge an interaction with the intention to edit its message later.

    This is valid for the following interaction types:

    * [`hikari.interactions.base_interactions.InteractionType.MESSAGE_COMPONENT`][]
    """

    MESSAGE_UPDATE = 7
    """An immediate interaction response with instructions on how to update its message.

    This is valid for the following interaction types:

    * [`hikari.interactions.base_interactions.InteractionType.MESSAGE_COMPONENT`][]
    """

    AUTOCOMPLETE = 8
    """Respond to an autocomplete interaction with suggested choices.

    This is valid for the following interaction types:

    * [`hikari.interactions.base_interactions.InteractionType.AUTOCOMPLETE`][]
    """

    MODAL = 9
    """An immediate interaction response with instructions to display a modal.

    This is valid for the following interaction types:

    * [`hikari.interactions.base_interactions.InteractionType.MODAL_SUBMIT`][]
    """

    PREMIUM_REQUIRED = 10
    """An immediate interaction response with a premium upsell button
    Only available for apps with monetization enabled.

    This is valid for the following interaction types:

    * `InteractionType.APPLICATION_COMMAND`
    * `InteractionType.MESSAGE_COMPONENT`
    * `InteractionType.MODAL_SUBMIT`
    """


MESSAGE_RESPONSE_TYPES: typing.Final[typing.AbstractSet[MessageResponseTypesT]] = frozenset(
    [ResponseType.MESSAGE_CREATE, ResponseType.MESSAGE_UPDATE]
)
"""Set of the response types which are valid for message responses.

This includes the following:

* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_CREATE`][]
* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_UPDATE`][]
"""

MessageResponseTypesT = typing.Literal[ResponseType.MESSAGE_CREATE, 4, ResponseType.MESSAGE_UPDATE, 7]
"""Type-hint of the response types which are valid for message responses.

The following are valid for this:

* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_CREATE`][]/`4`
* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_UPDATE`][]/`7`
"""

DEFERRED_RESPONSE_TYPES: typing.Final[typing.AbstractSet[DeferredResponseTypesT]] = frozenset(
    [ResponseType.DEFERRED_MESSAGE_CREATE, ResponseType.DEFERRED_MESSAGE_UPDATE]
)
"""Set of the response types which are valid for deferred messages responses.

This includes the following:

* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE`][]
* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE`][]
"""

DeferredResponseTypesT = typing.Literal[
    ResponseType.DEFERRED_MESSAGE_CREATE, 5, ResponseType.DEFERRED_MESSAGE_UPDATE, 6
]
"""Type-hint of the response types which are valid for deferred messages responses.

The following are valid for this:

* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE`][]/`5`
* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE`][]/`6`
"""


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class PartialInteraction(snowflakes.Unique, webhooks.ExecutableWebhook):
    """The base model for all interaction models."""

    app: traits.RESTAware = attrs.field(repr=False, eq=False, metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    """Client application that models may use for procedures."""

    id: snowflakes.Snowflake = attrs.field(hash=True, repr=True)
    # <<inherited docstring from Unique>>.

    application_id: snowflakes.Snowflake = attrs.field(eq=False, repr=False)
    """ID of the application this interaction belongs to."""

    type: InteractionType = attrs.field(eq=False, repr=True)
    """The type of interaction this is."""

    token: str = attrs.field(eq=False, repr=False)
    """The interaction's token."""

    version: int = attrs.field(eq=False, repr=True)
    """Version of the interaction system this interaction is under."""

    app_permissions: permissions_.Permissions | None = attrs.field(eq=False, hash=False, repr=False)
    """Permissions the bot has in this interaction's channel if it's in a guild."""

    user: users.User = attrs.field(eq=False, hash=False, repr=True)
    """The user who triggered this interaction."""

    member: InteractionMember | None = attrs.field(eq=False, hash=False, repr=True)
    """The member who triggered this interaction.

    This will be [`None`][] for interactions triggered in DMs.

    !!! note
        This member object comes with the extra field `permissions` which
        contains the member's permissions in the current channel.
    """

    channel: InteractionChannel = attrs.field(eq=False, repr=False)
    """The channel this interaction was triggered in."""

    guild_id: snowflakes.Snowflake | None = attrs.field(eq=False, hash=False, repr=True)
    """ID of the guild this modal interaction event was triggered in.

    This will be [`None`][] for modal interactions triggered in DMs.
    """

    guild_locale: str | locales.Locale | None = attrs.field(eq=False, hash=False, repr=True)
    """The preferred language of the guild this component interaction was triggered in.

    This will be [`None`][] for component interactions triggered in DMs.

    !!! note
        This value can usually only be changed if [COMMUNITY] is in [`hikari.guilds.Guild.features`][]
        for the guild and will otherwise default to `en-US`.
    """

    locale: str | locales.Locale = attrs.field(eq=False, hash=False, repr=True)
    """The selected language of the user who triggered this modal interaction."""

    authorizing_integration_owners: typing.Mapping[applications.ApplicationIntegrationType, snowflakes.Snowflake] = (
        attrs.field(eq=False, repr=True)
    )
    """A mapping of the [applications.ApplicationIntegrationType] to the related guild or user ID."""

    context: applications.ApplicationContextType = attrs.field(eq=False, repr=True)
    """The interaction context."""

    entitlements: typing.Sequence[monetization.Entitlement] = attrs.field(eq=False, hash=False, repr=True)
    """For monetized apps, any entitlements for the invoking user, represents access to SKUs."""

    @property
    def channel_id(self) -> snowflakes.Snowflake:
        """The ID of the channel this interaction was invoked in."""
        return self.channel.id

    @property
    @typing_extensions.override
    def webhook_id(self) -> snowflakes.Snowflake:
        # <<inherited docstring from ExecutableWebhook>>.
        return self.application_id

    async def fetch_guild(self) -> guilds.RESTGuild | None:
        """Fetch the guild this interaction happened in.

        Returns
        -------
        typing.Optional[hikari.guilds.RESTGuild]
            Object of the guild this interaction happened in or [`None`][]
            if this occurred within a DM channel.

        Raises
        ------
        hikari.errors.ForbiddenError
            If you are not part of the guild.
        hikari.errors.NotFoundError
            If the guild is not found.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        if not self.guild_id:
            return None

        return await self.app.rest.fetch_guild(self.guild_id)

    def get_guild(self) -> guilds.GatewayGuild | None:
        """Get the object of the guild this interaction was triggered in from the cache.

        Returns
        -------
        typing.Optional[hikari.guilds.GatewayGuild]
            The object of the guild if found, else [`None`][].
        """
        if self.guild_id and isinstance(self.app, traits.CacheAware):
            return self.app.cache.get_guild(self.guild_id)

        return None


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class PartialInteractionMetadata:
    """Metadata about the interaction, including the source of the interaction and relevant server and user IDs."""

    interaction_id: snowflakes.Snowflake = attrs.field(hash=True, repr=True)
    """The ID for this message interaction."""

    type: InteractionType | int = attrs.field(eq=False, repr=True)
    """The type of this message interaction."""

    user: users.User = attrs.field(eq=False, repr=True)
    """The user who triggered the interaction."""

    authorizing_integration_owners: typing.Mapping[applications.ApplicationIntegrationType, snowflakes.Snowflake] = (
        attrs.field(eq=False, repr=True)
    )
    """A mapping of the [applications.ApplicationIntegrationType] to the related guild or user ID."""

    original_response_message_id: snowflakes.Snowflake | None = attrs.field(hash=True, repr=True)
    """The ID of the original response message."""


class PremiumResponseMixin(PartialInteraction):
    """Mixin' class for all interaction types which can be responded to with a premium upsell."""

    __slots__: typing.Sequence[str] = ()

    async def create_premium_required_response(self) -> None:
        """Create a response by sending a premium upsell.

        Raises
        ------
        hikari.errors.ForbiddenError
            If you cannot access the target interaction.
        hikari.errors.NotFoundError
            If the initial response isn't found.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.create_premium_required_response(self, self.token)


class MessageResponseMixin(PartialInteraction, typing.Generic[_CommandResponseTypesT]):
    """Mixin' class for all interaction types which can be responded to with a message."""

    __slots__: typing.Sequence[str] = ()

    async def fetch_initial_response(self) -> messages.Message:
        """Fetch the initial response of this interaction.

        Returns
        -------
        hikari.messages.Message
            Message object of the initial response.

        Raises
        ------
        hikari.errors.ForbiddenError
            If you cannot access the target interaction.
        hikari.errors.NotFoundError
            If the initial response isn't found.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.fetch_interaction_response(self.application_id, self.token)

    async def create_initial_response(
        self,
        response_type: _CommandResponseTypesT,
        content: undefined.UndefinedOr[typing.Any] = undefined.UNDEFINED,
        *,
        flags: int | messages.MessageFlag | undefined.UndefinedType = undefined.UNDEFINED,
        tts: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        attachment: undefined.UndefinedNoneOr[files.Resourceish] = undefined.UNDEFINED,
        attachments: undefined.UndefinedNoneOr[typing.Sequence[files.Resourceish]] = undefined.UNDEFINED,
        component: undefined.UndefinedNoneOr[special_endpoints.ComponentBuilder] = undefined.UNDEFINED,
        components: undefined.UndefinedNoneOr[
            typing.Sequence[special_endpoints.ComponentBuilder]
        ] = undefined.UNDEFINED,
        embed: undefined.UndefinedNoneOr[embeds_.Embed] = undefined.UNDEFINED,
        embeds: undefined.UndefinedNoneOr[typing.Sequence[embeds_.Embed]] = undefined.UNDEFINED,
        poll: undefined.UndefinedOr[special_endpoints.PollBuilder] = undefined.UNDEFINED,
        mentions_everyone: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        user_mentions: undefined.UndefinedOr[
            snowflakes.SnowflakeishSequence[users.PartialUser] | bool
        ] = undefined.UNDEFINED,
        role_mentions: undefined.UndefinedOr[
            snowflakes.SnowflakeishSequence[guilds.PartialRole] | bool
        ] = undefined.UNDEFINED,
    ) -> None:
        """Create the initial response for this interaction.

        !!! warning
            Calling this on an interaction which already has an initial
            response will result in this raising a [`hikari.errors.NotFoundError`][].
            This includes if the REST interaction server has already responded
            to the request.

        Parameters
        ----------
        response_type
            The type of interaction response this is.
        content
            If provided, the message contents. If
            [`hikari.undefined.UNDEFINED`][], then nothing will be sent
            in the content. Any other value here will be cast to a
            [`str`][].

            If this is a [`hikari.embeds.Embed`][] and no `embed` nor `embeds` kwarg
            is provided, then this will instead update the embed. This allows
            for simpler syntax when sending an embed alone.
        attachment
            If provided, the message attachment. This can be a resource,
            or string of a path on your computer or a URL.
        attachments
            If provided, the message attachments. These can be resources, or
            strings consisting of paths on your computer or URLs.
        component
            If provided, builder object of the component to include in this message.
        components
            If provided, a sequence of the component builder objects to include
            in this message.
        embed
            If provided, the message embed.
        embeds
            If provided, the message embeds.
        poll
            If provided, the poll to set on the message.
        flags
            If provided, the message flags this response should have.

            As of writing the only message flags which can be set here are
            [`hikari.messages.MessageFlag.EPHEMERAL`][],
            [`hikari.messages.MessageFlag.SUPPRESS_NOTIFICATIONS`][]
            and [`hikari.messages.MessageFlag.SUPPRESS_EMBEDS`][].
        tts
            If provided, whether the message will be read out by a screen
            reader using Discord's TTS (text-to-speech) system.
        mentions_everyone
            If provided, whether the message should parse @everyone/@here
            mentions.
        user_mentions
            If provided, and [`True`][], all user mentions will be detected.
            If provided, and [`False`][], all user mentions will be ignored
            if appearing in the message body.
            Alternatively this may be a collection of
            [`hikari.snowflakes.Snowflake`][], or
            [`hikari.users.PartialUser`][] derivatives to enforce mentioning
            specific users.
        role_mentions
            If provided, and [`True`][], all role mentions will be detected.
            If provided, and [`False`][], all role mentions will be ignored
            if appearing in the message body.
            Alternatively this may be a collection of
            [`hikari.snowflakes.Snowflake`][], or
            [`hikari.guilds.PartialRole`][] derivatives to enforce mentioning
            specific roles.

        Raises
        ------
        ValueError
            If more than 100 unique objects/entities are passed for
            `role_mentions` or `user_mentions`.
        TypeError
            If both `embed` and `embeds` are specified.
        hikari.errors.BadRequestError
            This may be raised in several discrete situations, such as messages
            being empty with no embeds; messages with more than
            2000 characters in them, embeds that exceed one of the many embed
            limits; invalid image URLs in embeds.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.NotFoundError
            If the interaction is not found or if the interaction's initial
            response has already been created.
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        await self.app.rest.create_interaction_response(
            self.id,
            self.token,
            response_type,
            content,
            tts=tts,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            embed=embed,
            embeds=embeds,
            poll=poll,
            flags=flags,
            mentions_everyone=mentions_everyone,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
        )

    async def edit_initial_response(
        self,
        content: undefined.UndefinedNoneOr[typing.Any] = undefined.UNDEFINED,
        *,
        attachment: undefined.UndefinedNoneOr[files.Resourceish | messages.Attachment] = undefined.UNDEFINED,
        attachments: undefined.UndefinedNoneOr[
            typing.Sequence[files.Resourceish | messages.Attachment]
        ] = undefined.UNDEFINED,
        component: undefined.UndefinedNoneOr[special_endpoints.ComponentBuilder] = undefined.UNDEFINED,
        components: undefined.UndefinedNoneOr[
            typing.Sequence[special_endpoints.ComponentBuilder]
        ] = undefined.UNDEFINED,
        embed: undefined.UndefinedNoneOr[embeds_.Embed] = undefined.UNDEFINED,
        embeds: undefined.UndefinedNoneOr[typing.Sequence[embeds_.Embed]] = undefined.UNDEFINED,
        mentions_everyone: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        user_mentions: undefined.UndefinedOr[
            snowflakes.SnowflakeishSequence[users.PartialUser] | bool
        ] = undefined.UNDEFINED,
        role_mentions: undefined.UndefinedOr[
            snowflakes.SnowflakeishSequence[guilds.PartialRole] | bool
        ] = undefined.UNDEFINED,
    ) -> messages.Message:
        """Edit the initial response of this command interaction.

        !!! note
            Mentioning everyone, roles, or users in message edits currently
            will not send a push notification showing a new mention to people
            on Discord. It will still highlight in their chat as if they
            were mentioned, however.

        !!! warning
            If you specify a text `content`, `mentions_everyone`,
            `mentions_reply`, `user_mentions`, and `role_mentions` will default
            to [`False`][] as the message will be re-parsed for mentions. This will
            also occur if only one of the four are specified

            This is a limitation of Discord's design. If in doubt, specify all
            four of them each time.

        Parameters
        ----------
        content
            If provided, the message contents. If
            [`hikari.undefined.UNDEFINED`][], then nothing will be sent
            in the content. Any other value here will be cast to a
            [`str`][].

            If this is a [`hikari.embeds.Embed`][] and neither the
            `embed` or `embeds` kwargs are provided or if this is a
            [`hikari.files.Resourceish`][] and neither the `attachment` or
            `attachments` kwargs are provided, the values will be overwritten.
            This allows for simpler syntax when sending an embed or an
            attachment alone.

            Likewise, if this is a [`hikari.files.Resource`][], then the
            content is instead treated as an attachment if no `attachment` and
            no `attachments` kwargs are provided.
        attachment
            If provided, the attachment to set on the message. If
            [`hikari.undefined.UNDEFINED`][], the previous attachment, if
            present, is not changed. If this is [`None`][], then the
            attachment is removed, if present. Otherwise, the new attachment
            that was provided will be attached.
        attachments
            If provided, the attachments to set on the message. If
            [`hikari.undefined.UNDEFINED`][], the previous attachments, if
            present, are not changed. If this is [`None`][], then the
            attachments is removed, if present. Otherwise, the new attachments
            that were provided will be attached.
        component
            If provided, builder object of the component to set for this message.
            This component will replace any previously set components and passing
            [`None`][] will remove all components.
        components
            If provided, a sequence of the component builder objects set for
            this message. These components will replace any previously set
            components and passing [`None`][] or an empty sequence will
            remove all components.
        embed
            If provided, the embed to set on the message. If
            [`hikari.undefined.UNDEFINED`][], the previous embed(s) are not changed.
            If this is [`None`][] then any present embeds are removed.
            Otherwise, the new embed that was provided will be used as the
            replacement.
        embeds
            If provided, the embeds to set on the message. If
            [`hikari.undefined.UNDEFINED`][], the previous embed(s) are not changed.
            If this is [`None`][] then any present embeds are removed.
            Otherwise, the new embeds that were provided will be used as the
            replacement.
        mentions_everyone
            If provided, whether the message should parse @everyone/@here
            mentions.
        user_mentions
            If provided, and [`True`][], all user mentions will be detected.
            If provided, and [`False`][], all user mentions will be ignored
            if appearing in the message body.
            Alternatively this may be a collection of
            [`hikari.snowflakes.Snowflake`][], or
            [`hikari.users.PartialUser`][] derivatives to enforce mentioning
            specific users.
        role_mentions
            If provided, and [`True`][], all role mentions will be detected.
            If provided, and [`False`][], all role mentions will be ignored
            if appearing in the message body.
            Alternatively this may be a collection of
            [`hikari.snowflakes.Snowflake`][], or
            [`hikari.guilds.PartialRole`][] derivatives to enforce mentioning
            specific roles.

        Returns
        -------
        hikari.messages.Message
            The edited message.

        Raises
        ------
        ValueError
            If more than 100 unique objects/entities are passed for
            `role_mentions` or `user_mentions`.
        TypeError
            If both `embed` and `embeds` are specified.
        hikari.errors.BadRequestError
            This may be raised in several discrete situations, such as messages
            being empty with no attachments or embeds; messages with more than
            2000 characters in them, embeds that exceed one of the many embed
            limits; too many attachments; attachments that are too large;
            invalid image URLs in embeds; too many components.
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.NotFoundError
            If the interaction or the message are not found.
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        return await self.app.rest.edit_interaction_response(
            self.application_id,
            self.token,
            content,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            embed=embed,
            embeds=embeds,
            mentions_everyone=mentions_everyone,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
        )

    async def delete_initial_response(self) -> None:
        """Delete the initial response of this interaction.

        Raises
        ------
        hikari.errors.UnauthorizedError
            If you are unauthorized to make the request (invalid/missing token).
        hikari.errors.NotFoundError
            If the interaction or response is not found.
        hikari.errors.RateLimitTooLongError
            Raised in the event that a rate limit occurs that is
            longer than `max_rate_limit` when making a request.
        hikari.errors.InternalServerError
            If an internal error occurs on Discord while handling the request.
        """
        await self.app.rest.delete_interaction_response(self.application_id, self.token)


class ModalResponseMixin(PartialInteraction):
    """Mixin' class for all interaction types which can be responded to with a modal."""

    __slots__: typing.Sequence[str] = ()

    async def create_modal_response(
        self,
        title: str,
        custom_id: str,
        component: undefined.UndefinedOr[special_endpoints.ComponentBuilder] = undefined.UNDEFINED,
        components: undefined.UndefinedOr[typing.Sequence[special_endpoints.ComponentBuilder]] = undefined.UNDEFINED,
    ) -> None:
        """Create a response by sending a modal.

        Parameters
        ----------
        title
            The title that will show up in the modal.
        custom_id
            Developer set custom ID used for identifying interactions with this modal.
        component
            A component builder to send in this modal.
        components
            A sequence of component builders to send in this modal.

        Raises
        ------
        ValueError
            If both `component` and `components` are specified or if none are specified.
        """
        await self.app.rest.create_modal_response(
            self.id, self.token, title=title, custom_id=custom_id, component=component, components=components
        )

    def build_modal_response(self, title: str, custom_id: str) -> special_endpoints.InteractionModalBuilder:
        """Create a builder for a modal interaction response.

        Parameters
        ----------
        title
            The title that will show up in the modal.
        custom_id
            Developer set custom ID used for identifying interactions with this modal.

        Returns
        -------
        hikari.api.special_endpoints.InteractionModalBuilder
            The interaction modal response builder object.
        """
        return self.app.rest.interaction_modal_builder(title=title, custom_id=custom_id)


@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class InteractionMember(guilds.Member):
    """Model of the member who triggered an interaction.

    Unlike [`hikari.guilds.Member`][], this object comes with an extra
    [`hikari.interactions.base_interactions.InteractionMember.permissions`][] field.
    """

    permissions: permissions_.Permissions = attrs.field(eq=False, hash=False, repr=False)
    """Permissions the member has in the current channel."""


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class InteractionChannel(channels.PartialChannel):
    """Represents partial channels returned as resolved entities on interactions."""

    permissions: permissions_.Permissions = attrs.field(eq=False, hash=False, repr=True)
    """Permissions the command's executor has in this channel."""

    parent_id: snowflakes.Snowflake | None = attrs.field(eq=False, hash=False, repr=True)
    """The parent ID of the channel.

    This will be [`None`][] for DM channels and guild channels that have no parent.
    """

    thread_metadata: channels.ThreadMetadata | None = attrs.field(eq=False, hash=False, repr=False)
    """The thread metadata, if the channel is a thread."""


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class ResolvedOptionData:
    """Represents the resolved objects of entities referenced in a command's options."""

    attachments: typing.Mapping[snowflakes.Snowflake, messages.Attachment] = attrs.field(repr=False)
    """Mapping of snowflake IDs to the attachment objects."""

    channels: typing.Mapping[snowflakes.Snowflake, InteractionChannel] = attrs.field(repr=False)
    """Mapping of snowflake IDs to the resolved option partial channel objects."""

    members: typing.Mapping[snowflakes.Snowflake, InteractionMember] = attrs.field(repr=False)
    """Mapping of snowflake IDs to the resolved option member objects."""

    messages: typing.Mapping[snowflakes.Snowflake, messages.Message] = attrs.field(repr=False)
    """Mapping of snowflake IDs to the resolved option partial message objects."""

    roles: typing.Mapping[snowflakes.Snowflake, guilds.Role] = attrs.field(repr=False)
    """Mapping of snowflake IDs to the resolved option role objects."""

    users: typing.Mapping[snowflakes.Snowflake, users.User] = attrs.field(repr=False)
    """Mapping of snowflake IDs to the resolved option user objects."""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Core app interface for application implementations."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "CacheAware",
    "EntityFactoryAware",
    "EventFactoryAware",
    "EventManagerAware",
    "ExecutorAware",
    "GatewayBotAware",
    "IntentsAware",
    "InteractionServerAware",
    "NetworkSettingsAware",
    "RESTAware",
    "RESTBotAware",
    "Runnable",
    "ShardAware",
    "VoiceAware",
)

import abc
import typing

from hikari import presences
from hikari import undefined
from hikari.internal import fast_protocol

if typing.TYPE_CHECKING:
    import datetime
    from concurrent import futures

    from typing_extensions import Self

    from hikari import channels
    from hikari import guilds
    from hikari import intents as intents_
    from hikari import snowflakes
    from hikari import users as users_
    from hikari.api import cache as cache_
    from hikari.api import config
    from hikari.api import entity_factory as entity_factory_
    from hikari.api import event_factory as event_factory_
    from hikari.api import event_manager as event_manager_
    from hikari.api import interaction_server as interaction_server_
    from hikari.api import rest as rest_
    from hikari.api import shard as gateway_shard
    from hikari.api import voice as voice_


@typing.runtime_checkable
class NetworkSettingsAware(fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for any component aware of network settings."""

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def http_settings(self) -> config.HTTPSettings:
        """HTTP settings in use by this component."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def proxy_settings(self) -> config.ProxySettings:
        """Proxy settings in use by this component."""
        raise NotImplementedError


@typing.runtime_checkable
class EventManagerAware(fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for a event manager-aware object.

    event manager-aware components are able to manage event listeners and waiters.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def event_manager(self) -> event_manager_.EventManager:
        """Event manager for this object."""
        raise NotImplementedError


@typing.runtime_checkable
class EntityFactoryAware(fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for an entity factory-aware object.

    These components will be able to construct library entities.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def entity_factory(self) -> entity_factory_.EntityFactory:
        """Entity factory implementation for this object."""
        raise NotImplementedError


@typing.runtime_checkable
class ExecutorAware(fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for an executor-aware object.

    These components will contain an `executor` attribute that may return
    a [`concurrent.futures.Executor`][] or [`None`][] if the
    default [`asyncio`][] thread pool for the event loop is used.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def executor(self) -> futures.Executor | None:
        """Executor to use for blocking operations.

        This may return [`None`][] if the default [`asyncio`][] thread pool
        should be used instead.
        """
        raise NotImplementedError


@typing.runtime_checkable
class EventFactoryAware(fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for an event factory-aware object.

    These components are able to construct library events.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def event_factory(self) -> event_factory_.EventFactory:
        """Event factory component."""
        raise NotImplementedError


@typing.runtime_checkable
class IntentsAware(fast_protocol.FastProtocolChecking, typing.Protocol):
    """A component that is aware of the application intents."""

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def intents(self) -> intents_.Intents:
        """Intents registered for the application."""
        raise NotImplementedError


@typing.runtime_checkable
class RESTAware(
    EntityFactoryAware, NetworkSettingsAware, ExecutorAware, fast_protocol.FastProtocolChecking, typing.Protocol
):
    """Structural supertype for a REST-aware object.

    These are able to perform REST API calls.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def rest(self) -> rest_.RESTClient:
        """REST client to use for HTTP requests."""
        raise NotImplementedError


@typing.runtime_checkable
class VoiceAware(fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for a voice-aware object.

    This is an object that provides a `voice` property to allow the creation
    of custom voice client instances.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def voice(self) -> voice_.VoiceComponent:
        """Voice connection manager component for this application."""
        raise NotImplementedError


@typing.runtime_checkable
class ShardAware(
    IntentsAware, NetworkSettingsAware, ExecutorAware, VoiceAware, fast_protocol.FastProtocolChecking, typing.Protocol
):
    """Structural supertype for a shard-aware object.

    These will expose a mapping of shards, the intents in use
    and the bot user object.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def heartbeat_latencies(self) -> typing.Mapping[int, float]:
        """Mapping of shard ID to heartbeat latency.

        Any shards that are not yet started will be `float('nan')`.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def heartbeat_latency(self) -> float:
        """Average heartbeat latency of all started shards.

        If no shards are started, this will return `float('nan')`.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def shards(self) -> typing.Mapping[int, gateway_shard.GatewayShard]:
        """Mapping of shards in this application instance.

        Each shard ID is mapped to the corresponding shard instance.

        If the application has not started, it is acceptable to assume the
        result of this call will be an empty mapping.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def shard_count(self) -> int:
        """Number of shards in the total application.

        This may not be the same as the size of `shards`. If the application
        is auto-sharded, this may be `0` until the shards are started.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_me(self) -> users_.OwnUser | None:
        """Return the bot user, if known.

        This should be available as soon as the bot has fired the
        [`hikari.events.lifetime_events.StartingEvent`][].

        Until then, this may or may not be [`None`][].

        Returns
        -------
        typing.Optional[hikari.users.OwnUser]
            The bot user, if known, otherwise [`None`][].
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def update_presence(
        self,
        *,
        idle_since: undefined.UndefinedNoneOr[datetime.datetime] = undefined.UNDEFINED,
        status: undefined.UndefinedOr[presences.Status] = undefined.UNDEFINED,
        activity: undefined.UndefinedNoneOr[presences.Activity] = undefined.UNDEFINED,
        afk: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
    ) -> None:
        """Update the presence on all shards.

        This call will patch the presence on each shard. This means that
        unless you explicitly specify a parameter, the previous value will be
        retained. This means you do not have to track the global presence
        in your code.

        !!! note
            This will only send the update payloads to shards that are alive.
            Any shards that are not alive will cache the new presence for
            when they do start.

        !!! note
            If you want to set presences per shard, access the shard you wish
            to update (e.g. by using [`hikari.GatewayBot.shards`][]), and call
            [`hikari.api.shard.GatewayShard.update_presence`][] on that shard.
            This method is simply a facade to make performing this in bulk
            simpler.

        Parameters
        ----------
        idle_since
            The datetime that the user started being idle. If undefined, this
            will not be changed.
        afk
            Whether to be marked as AFK. If undefined, this will not be
            changed.
        activity
            The activity to appear to be playing. If undefined, this will not be
            changed.
        status
            The web status to show. If undefined, this will not be changed.

        Raises
        ------
        hikari.errors.ComponentStateConflictError
            When the shard is not connected so it cannot be interacted with.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def update_voice_state(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        channel: snowflakes.SnowflakeishOr[channels.GuildVoiceChannel] | None,
        *,
        self_mute: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        self_deaf: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
    ) -> None:
        """Update the voice state for this bot in a given guild.

        Parameters
        ----------
        guild
            The guild or guild ID to update the voice state for.
        channel
            The channel or channel ID to update the voice state for. If [`None`][]
            then the bot will leave the voice channel that it is in for the
            given guild.
        self_mute
            If specified and [`True`][], the bot will mute itself in that
            voice channel. If [`False`][], then it will unmute itself.
        self_deaf
            If specified and [`True`][], the bot will deafen itself in that
            voice channel. If [`False`][], then it will undeafen itself.

        Raises
        ------
        RuntimeError
            If the guild passed isn't covered by any of the shards in this sharded
            client.
        hikari.errors.ComponentStateConflictError
            When the shard is not connected so it cannot be interacted with.
        """

    @abc.abstractmethod
    async def request_guild_members(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        *,
        include_presences: undefined.UndefinedOr[bool] = undefined.UNDEFINED,
        query: str = "",
        limit: int = 0,
        users: undefined.UndefinedOr[snowflakes.SnowflakeishSequence[users_.User]] = undefined.UNDEFINED,
        nonce: undefined.UndefinedOr[str] = undefined.UNDEFINED,
    ) -> None:
        """Request for a guild chunk.

        !!! note
            To request the full list of members, set `query` to `""` (empty
            string) and `limit` to `0`.

        Parameters
        ----------
        guild
            The guild to request chunk for.
        include_presences
            If provided, whether to request presences.
        query
            If not `""`, request the members which username starts with the string.
        limit
            Maximum number of members to send matching the query.
        users
            If provided, the users to request for.
        nonce
            If provided, the nonce to be sent with guild chunks.

        Raises
        ------
        ValueError
            If trying to specify `users` with `query`/`limit`, if `limit` is not between
            0 and 100, both inclusive or if `users` length is over 100.
        hikari.errors.MissingIntentError
            When trying to request presences without the [`hikari.intents.Intents.GUILD_MEMBERS`][] or when trying to
            request the full list of members without [`hikari.intents.Intents.GUILD_PRESENCES`][].
        RuntimeError
            If the guild passed isn't covered by any of the shards in this sharded
            client.
        hikari.errors.ComponentStateConflictError
            When the shard is not connected so it cannot be interacted with.
        """


@typing.runtime_checkable
class InteractionServerAware(RESTAware, EntityFactoryAware, fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for a interaction REST server-aware object."""

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def interaction_server(self) -> interaction_server_.InteractionServer:
        """Interaction server this app is bound to."""
        raise NotImplementedError


@typing.runtime_checkable
class CacheAware(fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for a cache-aware object.

    Any cache-aware objects are able to access the Discord application cache.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def cache(self) -> cache_.Cache:
        """Immutable cache implementation for this object."""
        raise NotImplementedError


@typing.runtime_checkable
class Runnable(fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural super-type for an application which can be run independently."""

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def is_alive(self) -> bool:
        """Whether the application is running or not.

        This is useful as some functions might raise
        [`hikari.errors.ComponentStateConflictError`][] if this is
        [`False`][].
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def close(self) -> None:
        """Kill the application by shutting all components down."""

    @abc.abstractmethod
    async def join(self) -> None:
        """Wait indefinitely until the application closes.

        This can be placed in a task and cancelled without affecting the
        application runtime itself. Any exceptions raised by shards will be
        propagated to here.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run(self) -> None:
        """Start the application and block until it's finished running."""
        raise NotImplementedError

    @abc.abstractmethod
    async def start(self) -> None:
        """Start the application and then return."""
        raise NotImplementedError


@typing.runtime_checkable
class GatewayBotAware(
    RESTAware,
    Runnable,
    ShardAware,
    EventFactoryAware,
    EventManagerAware,
    CacheAware,
    fast_protocol.FastProtocolChecking,
    typing.Protocol,
):
    """Structural supertype for a component that has all the gateway components."""

    __slots__: typing.Sequence[str] = ()


@typing.runtime_checkable
class RESTBotAware(InteractionServerAware, Runnable, fast_protocol.FastProtocolChecking, typing.Protocol):
    """Structural supertype for a component that has all the RESTful components."""

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def on_shutdown(self) -> typing.Sequence[typing.Callable[[Self], typing.Coroutine[typing.Any, typing.Any, None]]]:
        """Sequence of the bot's asynchronous shutdown callbacks."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def on_startup(self) -> typing.Sequence[typing.Callable[[Self], typing.Coroutine[typing.Any, typing.Any, None]]]:
        """Sequence of the bot's asynchronous startup callbacks."""
        raise NotImplementedError

    @abc.abstractmethod
    def add_shutdown_callback(
        self, callback: typing.Callable[[RESTBotAware], typing.Coroutine[typing.Any, typing.Any, None]], /
    ) -> None:
        """Add an asynchronous callback to be called when the bot shuts down.

        Parameters
        ----------
        callback
            The asynchronous shutdown callback to add.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def remove_shutdown_callback(
        self, callback: typing.Callable[[RESTBotAware], typing.Coroutine[typing.Any, typing.Any, None]], /
    ) -> None:
        """Remove an asynchronous shutdown callback from the bot.

        Parameters
        ----------
        callback
            The shutdown callback to remove.

        Raises
        ------
        ValueError
            If the callback was not registered.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def add_startup_callback(
        self, callback: typing.Callable[[RESTBotAware], typing.Coroutine[typing.Any, typing.Any, None]], /
    ) -> None:
        """Add an asynchronous callback to be called when the bot starts up.

        Parameters
        ----------
        callback
            The asynchronous startup callback to add.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def remove_startup_callback(
        self, callback: typing.Callable[[RESTBotAware], typing.Coroutine[typing.Any, typing.Any, None]], /
    ) -> None:
        """Remove an asynchronous startup callback from the bot.

        Parameters
        ----------
        callback
            The asynchronous startup callback to remove.

        Raises
        ------
        ValueError
            If the callback was not registered.
        """
        raise NotImplementedError



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import files
from hikari import presences
from hikari import snowflakes
from hikari import urls
from hikari.impl import gateway_bot
from hikari.internal import routes


@pytest.fixture
def mock_app():
    return mock.Mock(spec_set=gateway_bot.GatewayBot)


class TestActivityAssets:
    def test_make_large_asset_url_format_set_to_deprecated_ext_argument_if_provided(self):
        asset = presences.ActivityAssets(
            application_id=snowflakes.Snowflake(1),
            large_image="abc",
            large_text="def",
            small_image="ghi",
            small_text="jkl",
        )

        with mock.patch.object(
            routes, "CDN_APPLICATION_ASSET", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert asset.make_large_image_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, application_id=1, hash="abc", size=4096, file_format="JPEG", lossless=True
        )

    def test_large_image_url_property(self):
        asset = presences.ActivityAssets(
            application_id=None, large_image=None, large_text=None, small_image=None, small_text=None
        )

        with mock.patch.object(presences.ActivityAssets, "make_large_image_url") as make_large_image_url:
            result = asset.large_image_url

        assert result is make_large_image_url.return_value
        make_large_image_url.assert_called_once_with()

    def test_large_image_url_property_when_runtime_error(self):
        asset = presences.ActivityAssets(
            application_id=None, large_image=None, large_text=None, small_image=None, small_text=None
        )

        with mock.patch.object(
            presences.ActivityAssets, "make_large_image_url", side_effect=RuntimeError
        ) as make_large_image_url:
            result = asset.large_image_url

        assert result is None
        make_large_image_url.assert_called_once_with()

    def test_make_large_image_url(self):
        asset = presences.ActivityAssets(
            application_id=45123123, large_image="541sdfasdasd", large_text=None, small_image=None, small_text=None
        )

        with mock.patch.object(routes, "CDN_APPLICATION_ASSET") as route:
            assert asset.make_large_image_url(ext="FA", size=3121) is route.compile_to_file.return_value

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, application_id=45123123, hash="541sdfasdasd", size=3121, file_format="FA", lossless=True
        )

    def test_make_large_image_url_when_no_hash(self):
        asset = presences.ActivityAssets(
            application_id=None, large_image=None, large_text=None, small_image=None, small_text=None
        )

        assert asset.make_large_image_url() is None

    @pytest.mark.parametrize(
        ("asset_hash", "expected"), [("mp:541sdfasdasd", "https://media.discordapp.net/541sdfasdasd")]
    )
    def test_make_large_image_url_when_dynamic_url(self, asset_hash: str, expected: str):
        asset = presences.ActivityAssets(
            application_id=None, large_image=asset_hash, large_text=None, small_image=None, small_text=None
        )

        assert asset.make_large_image_url() == files.URL(expected)

    def test_make_large_image_url_when_unknown_dynamic_url(self):
        asset = presences.ActivityAssets(
            application_id=None, large_image="uwu:nou", large_text=None, small_image=None, small_text=None
        )

        with pytest.raises(RuntimeError, match="Unknown asset type"):
            asset.make_large_image_url()

    def test_make_small_asset_url_format_set_to_deprecated_ext_argument_if_provided(self):
        asset = presences.ActivityAssets(
            application_id=snowflakes.Snowflake(1),
            large_image="abc",
            large_text="def",
            small_image="ghi",
            small_text="jkl",
        )

        with mock.patch.object(
            routes, "CDN_APPLICATION_ASSET", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert asset.make_small_image_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, application_id=1, hash="ghi", size=4096, file_format="JPEG", lossless=True
        )

    def test_small_image_url_property(self):
        asset = presences.ActivityAssets(
            application_id=None, large_image=None, large_text=None, small_image=None, small_text=None
        )

        with mock.patch.object(presences.ActivityAssets, "make_small_image_url") as make_small_image_url:
            result = asset.small_image_url

        assert result is make_small_image_url.return_value
        make_small_image_url.assert_called_once_with()

    def test_small_image_url_property_when_runtime_error(self):
        asset = presences.ActivityAssets(
            application_id=None, large_image=None, large_text=None, small_image=None, small_text=None
        )

        with mock.patch.object(
            presences.ActivityAssets, "make_small_image_url", side_effect=RuntimeError
        ) as make_small_image_url:
            result = asset.small_image_url

        assert result is None
        make_small_image_url.assert_called_once_with()

    def test_make_small_image_url(self):
        asset = presences.ActivityAssets(
            application_id=123321, large_image=None, large_text=None, small_image="aseqwsdas", small_text=None
        )

        with mock.patch.object(routes, "CDN_APPLICATION_ASSET") as route:
            assert asset.make_small_image_url(ext="EAT", size=123312) is route.compile_to_file.return_value

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, application_id=123321, hash="aseqwsdas", size=123312, file_format="EAT", lossless=True
        )

    def test_make_small_image_url_when_no_hash(self):
        asset = presences.ActivityAssets(
            application_id=None, large_image=None, large_text=None, small_image=None, small_text=None
        )

        assert asset.make_small_image_url() is None

    @pytest.mark.parametrize(("asset_hash", "expected"), [("mp:4123fdssdf", "https://media.discordapp.net/4123fdssdf")])
    def test_make_small_image_url_when_dynamic_url(self, asset_hash: str, expected: str):
        asset = presences.ActivityAssets(
            application_id=None, large_image=None, large_text=None, small_image=asset_hash, small_text=None
        )

        assert asset.make_small_image_url() == files.URL(expected)

    def test_make_small_image_url_when_unknown_dynamic_url(self):
        asset = presences.ActivityAssets(
            application_id=None, large_image=None, large_text=None, small_image="meow:nyaa", small_text=None
        )

        with pytest.raises(RuntimeError, match="Unknown asset type"):
            asset.make_small_image_url()


class TestActivity:
    def test_str_operator(self):
        activity = presences.Activity(name="something", type=presences.ActivityType(1))
        assert str(activity) == "something"


class TestMemberPresence:
    @pytest.fixture
    def model(self, mock_app):
        return presences.MemberPresence(
            app=mock_app,
            user_id=snowflakes.Snowflake(432),
            guild_id=snowflakes.Snowflake(234),
            visible_status=presences.Status.ONLINE,
            activities=mock.Mock(presences.RichActivity),
            client_status=mock.Mock(presences.ClientStatus),
        )

    @pytest.mark.asyncio
    async def test_fetch_user(self, model):
        model.app.rest.fetch_user = mock.AsyncMock()

        assert await model.fetch_user() is model.app.rest.fetch_user.return_value
        model.app.rest.fetch_user.assert_awaited_once_with(432)

    @pytest.mark.asyncio
    async def test_fetch_member(self, model):
        model.app.rest.fetch_member = mock.AsyncMock()

        assert await model.fetch_member() is model.app.rest.fetch_member.return_value
        model.app.rest.fetch_member.assert_awaited_once_with(234, 432)



# Copyright (c) 2020 Tomxey
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime

import mock
import pytest

from hikari import applications
from hikari import emojis
from hikari import guilds
from hikari import snowflakes
from hikari import users


def make_user(user_id):
    return users.UserImpl(
        app=mock.Mock(),
        id=snowflakes.Snowflake(user_id),
        discriminator="0001",
        username="testing",
        global_name=None,
        avatar_decoration=None,
        avatar_hash=None,
        banner_hash=None,
        accent_color=None,
        is_bot=False,
        is_system=False,
        flags=users.UserFlag.NONE,
    )


def make_team_member(user_id):
    user = make_user(user_id)
    return applications.TeamMember(
        membership_state=applications.TeamMembershipState.ACCEPTED,
        permissions="*",
        team_id=snowflakes.Snowflake(1234),
        user=user,
    )


def make_guild_member(user_id):
    user = make_user(user_id)
    return guilds.Member(
        user=user,
        guild_id=snowflakes.Snowflake(2233),
        role_ids=[],
        joined_at=datetime.datetime.now(),
        nickname=user.username,
        premium_since=None,
        guild_avatar_decoration=None,
        guild_avatar_hash="no",
        guild_banner_hash="yes",
        is_deaf=False,
        is_mute=False,
        is_pending=False,
        raw_communication_disabled_until=None,
        guild_flags=guilds.GuildMemberFlags.NONE,
    )


def make_unicode_emoji():
    return emojis.UnicodeEmoji("\N{OK HAND SIGN}")


def make_custom_emoji(emoji_id):
    return emojis.CustomEmoji(id=emoji_id, name="testing", is_animated=False)


def make_known_custom_emoji(emoji_id):
    return emojis.KnownCustomEmoji(
        app=mock.Mock(),
        id=emoji_id,
        name="testing",
        is_animated=False,
        guild_id=snowflakes.Snowflake(123),
        role_ids=[],
        user=None,
        is_colons_required=False,
        is_managed=True,
        is_available=False,
    )


@pytest.mark.parametrize(
    ("a", "b", "eq"),
    [
        (make_user(1), make_team_member(1), True),
        (make_user(1), make_team_member(2), False),
        (make_user(1), make_guild_member(1), True),
        (make_user(1), make_guild_member(2), False),
        (make_team_member(1), make_guild_member(1), True),
        (make_team_member(1), make_guild_member(2), False),
        (make_custom_emoji(1), make_known_custom_emoji(1), True),
        (make_custom_emoji(1), make_known_custom_emoji(2), False),
        (make_unicode_emoji(), make_custom_emoji(1), False),
        (make_unicode_emoji(), make_known_custom_emoji(2), False),
    ],
    ids=[
        "User == Team Member",
        "User != Team Member",
        "User == Guild Member",
        "User != Guild Member",
        "Team Member == Guild Member",
        "Team Member != Guild Member",
        "Custom Emoji == Known Custom Emoji",
        "Custom Emoji != Known Custom Emoji",
        "Unicode Emoji != Custom Emoji",
        "Unicode Emoji != Known Custom Emoji",
    ],
)
def test_comparison(a: object, b: object, eq: bool) -> None:
    if eq:
        assert a == b
        assert b == a
        assert not a != b
        assert not b != a

    else:
        assert a != b
        assert b != a
        assert not a == b
        assert not b == a



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import http
import inspect

import mock
import pytest

from hikari import errors
from hikari import intents


class TestShardCloseCode:
    @pytest.mark.parametrize(("code", "expected"), [(1000, True), (1001, True), (4000, False), (4014, False)])
    def test_is_standard_property(self, code, expected):
        assert errors.ShardCloseCode(code).is_standard is expected


class TestComponentStateConflictError:
    @pytest.fixture
    def error(self):
        return errors.ComponentStateConflictError("some reason")

    def test_str(self, error):
        assert str(error) == "some reason"


class TestUnrecognisedEntityError:
    @pytest.fixture
    def error(self):
        return errors.UnrecognisedEntityError("some reason")

    def test_str(self, error):
        assert str(error) == "some reason"


class TestGatewayError:
    @pytest.fixture
    def error(self):
        return errors.GatewayError("some reason")

    def test_str(self, error):
        assert str(error) == "some reason"


class TestGatewayServerClosedConnectionError:
    @pytest.fixture
    def error(self):
        return errors.GatewayServerClosedConnectionError("some reason", 123)

    def test_str(self, error):
        assert str(error) == "Server closed connection with code 123 (some reason)"


class TestHTTPResponseError:
    @pytest.fixture
    def error(self):
        return errors.HTTPResponseError(
            "https://some.url", http.HTTPStatus.BAD_REQUEST, {}, "raw body", "message", 12345
        )

    def test_str(self, error):
        assert str(error) == "Bad Request 400: (12345) 'message' for https://some.url"

    def test_str_when_int_status_code(self, error):
        error.status = 699
        assert str(error) == "Unknown Status 699: (12345) 'message' for https://some.url"

    def test_str_when_message_is_None(self, error):
        error.message = None
        assert str(error) == "Bad Request 400: (12345) 'raw body' for https://some.url"

    def test_str_when_code_is_zero(self, error):
        error.code = 0
        assert str(error) == "Bad Request 400: 'message' for https://some.url"

    def test_str_when_code_is_not_zero(self, error):
        error.code = 100
        assert str(error) == "Bad Request 400: (100) 'message' for https://some.url"


class TestBadRequestError:
    @pytest.fixture
    def error(self):
        return errors.BadRequestError(
            "https://some.url",
            http.HTTPStatus.BAD_REQUEST,
            {},
            "raw body",
            errors={
                "": [{"code": "012", "message": "test error"}],
                "components": {
                    "0": {
                        "_errors": [
                            {"code": "123", "message": "something went wrong"},
                            {"code": "456", "message": "but more things too!"},
                        ]
                    }
                },
                "attachments": {"1": {"_errors": [{"code": "789", "message": "at this point, all wrong!"}]}},
            },
        )

    def test_str(self, error):
        assert str(error) == inspect.cleandoc(
            """
            Bad Request 400: 'raw body' for https://some.url

            root:
             - test error

            components.0:
             - something went wrong
             - but more things too!

            attachments.1:
             - at this point, all wrong!
            """
        )

    def test_str_when_dump_error_errors(self, error):
        with mock.patch.object(errors, "_dump_errors", side_effect=KeyError):
            string = str(error)

        assert string == inspect.cleandoc(
            """
            Bad Request 400: 'raw body' for https://some.url

            {
              "": [
                {
                  "code": "012",
                  "message": "test error"
                }
              ],
              "components": {
                "0": {
                  "_errors": [
                    {
                      "code": "123",
                      "message": "something went wrong"
                    },
                    {
                      "code": "456",
                      "message": "but more things too!"
                    }
                  ]
                }
              },
              "attachments": {
                "1": {
                  "_errors": [
                    {
                      "code": "789",
                      "message": "at this point, all wrong!"
                    }
                  ]
                }
              }
            }
            """
        )

    def test_str_when_cached(self, error):
        error._cached_str = "ok"

        with mock.patch.object(errors, "_dump_errors") as dump_errors:
            assert str(error) == "ok"

        dump_errors.assert_not_called()

    def test_str_when_no_errors(self, error):
        error.errors = None

        with mock.patch.object(errors, "_dump_errors") as dump_errors:
            assert str(error) == "Bad Request 400: 'raw body' for https://some.url"

        dump_errors.assert_not_called()


class TestRateLimitTooLongError:
    @pytest.fixture
    def error(self):
        return errors.RateLimitTooLongError(
            route="some route", is_global=False, retry_after=0, max_retry_after=60, reset_at=0, limit=0, period=0
        )

    def test_remaining(self, error):
        assert error.remaining == 0

    def test_str(self, error):
        assert str(error) == (
            "The request has been rejected, as you would be waiting for more than "
            "the max retry-after (60) on route 'some route' [is_global=False]"
        )


class TestBulkDeleteError:
    @pytest.fixture
    def error(self):
        return errors.BulkDeleteError(range(10))

    def test_str(self, error):
        assert str(error) == "Error encountered when bulk deleting messages (10 messages deleted)"


class TestMissingIntentError:
    @pytest.fixture
    def error(self):
        return errors.MissingIntentError(intents.Intents.GUILD_MEMBERS | intents.Intents.GUILD_EMOJIS)

    def test_str(self, error):
        assert str(error) == "You are missing the following intent(s): GUILD_EMOJIS, GUILD_MEMBERS"
